__author__ = "Simon Nilsson"

import os
import tkinter.ttk as ttk
from copy import deepcopy
from tkinter import *

import pandas as pd
import PIL.Image
from PIL import ImageTk

import simba
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.pop_ups.clf_add_remove_print_pop_up import PoseResetterPopUp
from simba.ui.pop_ups.create_user_defined_pose_configuration_pop_up import \
    CreateUserDefinedPoseConfigurationPopUp
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, FolderSelect, hxtScrollbar)
from simba.utils.checks import check_if_dir_exists, check_str
from simba.utils.config_creator import ProjectConfigCreator
from simba.utils.enums import Formats, Keys, Links, Methods, Options, Paths
from simba.utils.errors import DuplicationError, MissingProjectConfigEntryError
from simba.utils.lookups import (get_body_part_configurations,
                                 get_bp_config_codes, get_icons_paths)
from simba.video_processors.video_processing import \
    extract_frames_from_all_videos_in_directory


class ProjectCreatorPopUp(PopUpMixin):
    """
    Mixin for GUI pop-up windows that accept user-inputs.
    """

    def __init__(self):
        self.main_frm = Toplevel()
        self.main_frm.minsize(750, 750)
        self.main_frm.wm_title("PROJECT CONFIGURATION")
        self.main_frm.columnconfigure(0, weight=1)
        self.main_frm.rowconfigure(0, weight=1)
        parent_tab = ttk.Notebook(hxtScrollbar(self.main_frm))
        self.btn_icons = get_icons_paths()
        for k in self.btn_icons.keys():
            self.btn_icons[k]["img"] = ImageTk.PhotoImage(
                image=PIL.Image.open(
                    os.path.join(
                        os.path.dirname(__file__), self.btn_icons[k]["icon_path"]
                    )
                )
            )
        self.create_project_tab = ttk.Frame(parent_tab)
        self.import_videos_tab = ttk.Frame(parent_tab)
        self.import_data_tab = ttk.Frame(parent_tab)
        self.extract_frms_tab = ttk.Frame(parent_tab)

        parent_tab.add(
            self.create_project_tab,
            text=f'{"[ Create project config ]": ^20s}',
            compound="left",
            image=self.btn_icons["create"]["img"],
        )
        parent_tab.add(
            self.import_videos_tab,
            text=f'{"[ Import videos ]": ^20s}',
            compound="left",
            image=self.btn_icons["video"]["img"],
        )
        parent_tab.add(
            self.import_data_tab,
            text=f'{"[ Import tracking data ]": ^20s}',
            compound="left",
            image=self.btn_icons["pose"]["img"],
        )
        parent_tab.add(
            self.extract_frms_tab,
            text=f'{"[ Extract frames ]": ^20s}',
            compound="left",
            image=self.btn_icons["frames"]["img"],
        )

        parent_tab.grid(row=0, column=0, sticky=NW)

        self.settings_frm = CreateLabelFrameWithIcon(
            parent=self.create_project_tab,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.CREATE_PROJECT.value,
        )
        self.general_settings_frm = LabelFrame(
            self.settings_frm,
            text="GENERAL PROJECT SETTINGS",
            fg="black",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            padx=5,
            pady=5,
        )

        self.project_dir_select = FolderSelect(
            self.general_settings_frm, "Project directory:", lblwidth="25"
        )
        self.project_name_eb = Entry_Box(
            self.general_settings_frm, "Project name:", labelwidth="25"
        )
        self.file_type_dropdown = DropDownMenu(
            self.general_settings_frm,
            "Workflow file type:",
            Options.WORKFLOW_FILE_TYPE_OPTIONS.value,
            "25",
        )
        self.file_type_dropdown.setChoices(
            choice=Options.WORKFLOW_FILE_TYPE_OPTIONS.value[0]
        )

        self.clf_name_entries = []
        self.ml_settings_frm = LabelFrame(
            self.settings_frm,
            text="MACHINE LEARNING SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            padx=5,
            pady=5,
        )
        self.clf_cnt = Entry_Box(
            self.ml_settings_frm,
            "Number of classifiers (behaviors): ",
            "25",
            validation="numeric",
        )
        add_clf_btn = Button(
            self.ml_settings_frm,
            text="<Add predictive classifier(s)>",
            fg="blue",
            command=lambda: self.create_entry_boxes_from_entrybox(
                count=self.clf_cnt.entry_get,
                parent=self.ml_settings_frm,
                current_entries=self.clf_name_entries,
            ),
        )

        self.animal_settings_frm = LabelFrame(
            self.settings_frm,
            text="ANIMAL SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.tracking_type_dropdown = DropDownMenu(
            self.animal_settings_frm,
            "Type of Tracking",
            Options.TRACKING_TYPE_OPTIONS.value,
            "25",
            com=self.update_body_part_dropdown,
        )
        self.tracking_type_dropdown.setChoices(Options.TRACKING_TYPE_OPTIONS.value[0])

        project_animal_cnt_path = os.path.join(
            os.path.dirname(simba.__file__), Paths.SIMBA_NO_ANIMALS_PATH.value
        )
        self.animal_count_lst = list(
            pd.read_csv(project_animal_cnt_path, header=None)[0]
        )
        self.bp_lu = get_body_part_configurations()
        self.bp_config_codes = get_bp_config_codes()
        self.classical_tracking_options = deepcopy(
            Options.CLASSICAL_TRACKING_OPTIONS.value
        )
        self.multi_tracking_options = deepcopy(
            Options.MULTI_ANIMAL_TRACKING_OPTIONS.value
        )
        self.three_dim_tracking_options = deepcopy(
            Options.THREE_DIM_TRACKING_OPTIONS.value
        )
        self.user_defined_options = [
            x
            for x in list(self.bp_lu.keys())
            if x not in self.classical_tracking_options
            and x not in self.multi_tracking_options
            and x not in self.three_dim_tracking_options
        ]
        for k in self.bp_lu.keys():
            self.bp_lu[k]["img"] = ImageTk.PhotoImage(
                file=os.path.join(os.path.dirname(__file__), self.bp_lu[k]["img_path"])
            )
        self.classical_tracking_option_dict = {
            k: self.bp_lu[k] for k in self.classical_tracking_options
        }
        self.multi_tracking_option_dict = {
            k: self.bp_lu[k] for k in self.multi_tracking_options
        }
        self.classical_tracking_options.append(Methods.CREATE_POSE_CONFIG.value)
        self.multi_tracking_options.append(Methods.CREATE_POSE_CONFIG.value)
        self.three_dim_tracking_options.append(Methods.CREATE_POSE_CONFIG.value)
        self.classical_tracking_options.extend(self.user_defined_options)
        self.multi_tracking_options.extend(self.user_defined_options)
        self.three_dim_tracking_options.extend(self.user_defined_options)
        self.selected_tracking_dropdown = DropDownMenu(
            self.animal_settings_frm,
            "Body-part config",
            Options.CLASSICAL_TRACKING_OPTIONS.value,
            "25",
            com=self.update_img,
        )
        self.selected_tracking_dropdown.setChoices(self.classical_tracking_options[0])
        self.img_lbl = Label(
            self.animal_settings_frm,
            image=self.bp_lu[self.classical_tracking_options[0]]["img"],
        )
        reset_btn = Button(
            self.animal_settings_frm,
            text="RESET USER DEFINED POSE-CONFIGS",
            fg="red",
            command=lambda: PoseResetterPopUp(),
        )
        run_frm = Frame(master=self.settings_frm)
        create_project_btn = Button(
            run_frm,
            text="CREATE PROJECT CONFIG",
            fg="navy",
            font=("Helvetica", 16, "bold"),
            command=lambda: self.run(),
        )
        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.general_settings_frm.grid(row=0, column=0, sticky=NW)
        self.project_dir_select.grid(row=0, column=0, sticky=NW)
        self.project_name_eb.grid(row=1, column=0, sticky=NW)
        self.file_type_dropdown.grid(row=2, column=0, sticky=NW)

        self.ml_settings_frm.grid(row=1, column=0, sticky=NW)
        self.clf_cnt.grid(row=0, column=0, sticky=NW)
        add_clf_btn.grid(row=1, column=0, sticky=NW)

        self.animal_settings_frm.grid(row=2, column=0, sticky=NW)
        self.tracking_type_dropdown.grid(row=0, column=0, sticky=NW)
        self.selected_tracking_dropdown.grid(row=1, column=0, sticky=NW)
        self.img_lbl.grid(row=2, column=0, sticky=NW)
        reset_btn.grid(row=0, column=1, sticky=NW)
        run_frm.grid(row=3, column=0, sticky=NW)
        create_project_btn.grid(row=0, column=0, sticky=NW)

        self.create_import_videos_menu(parent_frm=self.import_videos_tab)
        self.create_import_pose_menu(parent_frm=self.import_data_tab)

        extract_frames_frm = LabelFrame(
            self.extract_frms_tab,
            text="EXTRACT FRAMES INTO PROJECT",
            fg="black",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        extract_frames_note = Label(
            extract_frames_frm,
            text="Note: This is no longer needed for any of the parts of the SimBA pipeline.\n Caution: This extract all frames from all videos in project. \n and is computationally expensive if there is a lot of videos at high frame rates/resolution.",
        )
        extract_frames_btn = Button(
            extract_frames_frm, text="EXTRACT FRAMES", fg="blue", command=lambda: None
        )

        extract_frames_frm.grid(row=0, column=0, sticky=NW)
        extract_frames_note.grid(row=0, column=0, sticky=NW)
        extract_frames_btn.grid(row=1, column=0, sticky=NW)
        self.update_body_part_dropdown(Methods.CLASSIC_TRACKING.value)
        self.main_frm.mainloop()

    def update_body_part_dropdown(self, selected_value):
        self.selected_tracking_dropdown.destroy()
        if selected_value == Methods.MULTI_TRACKING.value:
            self.selected_tracking_dropdown = DropDownMenu(
                self.animal_settings_frm,
                "Body-part config",
                self.multi_tracking_options,
                "25",
                com=self.update_img,
            )
            self.selected_tracking_dropdown.setChoices(self.multi_tracking_options[0])
            self.selected_tracking_dropdown.grid(row=1, column=0, sticky=NW)
        elif selected_value == Methods.CLASSIC_TRACKING.value:
            self.selected_tracking_dropdown = DropDownMenu(
                self.animal_settings_frm,
                "Body-part config",
                self.classical_tracking_options,
                "25",
                com=self.update_img,
            )
            self.selected_tracking_dropdown.setChoices(
                self.classical_tracking_options[0]
            )
            self.selected_tracking_dropdown.grid(row=1, column=0, sticky=NW)
        elif selected_value == Methods.THREE_D_TRACKING.value:
            self.selected_tracking_dropdown = DropDownMenu(
                self.animal_settings_frm,
                "Body-part config",
                self.three_dim_tracking_options,
                "25",
                com=self.update_img,
            )
            self.selected_tracking_dropdown.setChoices(
                self.three_dim_tracking_options[0]
            )
            self.selected_tracking_dropdown.grid(row=1, column=0, sticky=NW)
        self.update_img(self.selected_tracking_dropdown.getChoices())

    def update_img(self, selected_value):
        if selected_value != Methods.CREATE_POSE_CONFIG.value:
            self.img_lbl.config(image=self.bp_lu[selected_value]["img"])
        else:
            _ = CreateUserDefinedPoseConfigurationPopUp(
                master=self.main_frm, project_config_class=ProjectCreatorPopUp
            )

    def extract_frames(self):
        if not hasattr(self, "config_path"):
            raise MissingProjectConfigEntryError(
                msg="Create PROJECT CONFIG before extracting frames"
            )
        video_dir = os.path.join(os.path.dirname(self.config_path), "videos")
        extract_frames_from_all_videos_in_directory(
            config_path=self.config_path, directory=video_dir
        )

    def run(self):
        project_dir = self.project_dir_select.folder_path
        check_if_dir_exists(in_dir=project_dir)
        project_name = self.project_name_eb.entry_get
        check_str(name="PROJECT NAME", value=project_name, allow_blank=False)
        target_list = []
        for number, entry_box in enumerate(self.clf_name_entries):
            target_list.append(entry_box.entry_get.strip())
        if len(list(set(target_list))) != len(self.clf_name_entries):
            raise DuplicationError(msg="All classifier names have to be unique")

        selected_config = self.selected_tracking_dropdown.getChoices()

        if selected_config in self.bp_config_codes.keys():
            config_code = self.bp_config_codes[selected_config]
        else:
            config_code = Methods.USER_DEFINED.value

        config_idx = None
        for cnt, k in enumerate(self.bp_lu.keys()):
            if k == selected_config:
                config_idx = cnt
        animal_cnt = self.animal_count_lst[config_idx]

        config_creator = ProjectConfigCreator(
            project_path=project_dir,
            project_name=project_name,
            target_list=target_list,
            pose_estimation_bp_cnt=config_code,
            body_part_config_idx=config_idx,
            animal_cnt=animal_cnt,
            file_type=self.file_type_dropdown.getChoices(),
        )

        self.config_path = config_creator.config_path
        self.create_import_pose_menu(parent_frm=self.import_data_tab)
        self.create_import_videos_menu(parent_frm=self.import_videos_tab)


# ProjectCreatorPopUp()



import os
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.roi_tools.ROI_feature_analyzer import ROIFeatureCreator
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box)
from simba.utils.enums import ConfigKey, Formats, Keys, Links
from simba.utils.errors import NoROIDataError


class AppendROIFeaturesByAnimalPopUp(ConfigReader, PopUpMixin):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        if not os.path.isfile(self.roi_coordinates_path):
            raise NoROIDataError(
                msg=f"SIMBA ERROR: No ROIs have been defined. Please define ROIs before appending ROI-based features (no data file found at path {self.roi_coordinates_path})",
                source=self.__class__.__name__,
            )
        PopUpMixin.__init__(
            self, title="APPEND ROI FEATURES: BY ANIMALS", size=(400, 400)
        )
        self.animal_cnt_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SELECT NUMBER OF ANIMALS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ROI_FEATURES.value,
        )
        self.animal_cnt_dropdown = DropDownMenu(
            self.animal_cnt_frm,
            "# of animals",
            list(range(1, self.animal_cnt + 1)),
            labelwidth=20,
        )
        self.animal_cnt_dropdown.setChoices(1)
        self.animal_cnt_confirm_btn = Button(
            self.animal_cnt_frm,
            text="Confirm",
            command=lambda: self.create_settings_frm(),
        )
        self.animal_cnt_frm.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)
        self.animal_cnt_confirm_btn.grid(row=0, column=1, sticky=NW)
        self.main_frm.mainloop()

    def create_settings_frm(self):
        if hasattr(self, "setting_frm"):
            self.setting_frm.destroy()
            self.body_part_frm.destroy()
        self.setting_frm = LabelFrame(
            self.main_frm, text="SETTINGS", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.choose_bp_frm(parent=self.setting_frm, bp_options=self.body_parts_lst)
        self.setting_frm.grid(row=1, column=0, sticky=NW)
        self.create_run_frm(run_function=self.run)

    def run(self):
        if self.config.has_section(ConfigKey.ROI_SETTINGS.value):
            self.config.remove_section(ConfigKey.ROI_SETTINGS.value)
        self.config.add_section(ConfigKey.ROI_SETTINGS.value)

        for bp_cnt, bp_dropdown in self.body_parts_dropdowns.items():
            self.config.set(
                ConfigKey.ROI_SETTINGS.value,
                f"animal_{bp_cnt+1}_bp",
                bp_dropdown.getChoices(),
            )
        with open(self.config_path, "w") as f:
            self.config.write(f)

        roi_feature_creator = ROIFeatureCreator(config_path=self.config_path)
        roi_feature_creator.run()
        roi_feature_creator.save()


# AppendROIFeaturesByAnimalPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini')


# AppendROIFeaturesByAnimalPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')



import itertools
import os
from typing import Dict, List, Optional, Tuple, Union

import networkx as nx
from pyvis.network import Network

from simba.utils.checks import (check_float, check_instance, check_int,
                                check_iterable_length, check_str,
                                check_valid_hex_color)
from simba.utils.data import create_color_palette, find_ranked_colors
from simba.utils.errors import CountError, InvalidInputError


class NetworkMixin(object):
    """
    Methods to create and analyze time-dependent graphs from pose-estimation data.

    When working with pose-estimation data for more than two animals - over extended periods - it can be beneficial to
    represent the data as a graph where the animals feature as nodes are their relationship strengths are represented as edges.

    When formatted as a graph, we can compute (i) how the relationships between animal pairs change across time and recordings,
    (ii) the relative importance's and hierarchies of individual animals within the group, or (iii) identify sub-groups with the network.

    The critical component determining the results is how edge weights are represented. These edge weight values could be the amount of time animal bounding
    boxes overlap each other, aggregate distances between the animals, or how much time animals engange in coordinated behaviors. These values can be computed
    through other methods within SimBA mixin methods.

    Very much wip and so far primarily depend on `networkx <https://github.com/networkx/networkx>`_.

    References
    ----------
    See below references for mature and reliable packages (12/2023):

    .. [1] `networkx <https://github.com/networkx/networkx>`_
    .. [2] `igraph <https://github.com/networkx/networkx>`_
    """

    def __init__(self):
        pass

    @staticmethod
    def create_graph(data: Dict[Tuple[str, str], float]) -> nx.Graph():
        """
        Create a single undirected graph with single edges from on dictionary.

        :param Dict[Tuple[str, str], float] data: A dictionary where keys are tuples representing node pairs and values are the corresponding edge weights.
        :returns nx.Graph: A networkx graph with nodes and edges defined by the input data.

        :example:
        >>> data = {('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5}
        >>> graph = NetworkMixin.create_graph(data=data)
        """

        check_instance(
            source=NetworkMixin.create_graph.__name__,
            instance=data,
            accepted_types=dict,
        )
        for k, v in data.items():
            check_instance(
                source=NetworkMixin.create_graph.__name__,
                instance=k,
                accepted_types=tuple,
            )
            check_iterable_length(
                source=f"{NetworkMixin.create_graph.__name__} {k}",
                val=len(k),
                exact_accepted_length=2,
            )
            check_instance(
                source=f"{NetworkMixin.create_graph.__name__} {v}",
                instance=v,
                accepted_types=(float, int),
            )

        G = nx.Graph()
        for node_names in data.keys():
            G.add_node(node_names[0])
            G.add_node(node_names[1])
        for node_names, edge_weight in data.items():
            G.add_edge(node_names[0], node_names[1], weight=edge_weight)
        return G

    @staticmethod
    def create_multigraph(data: Dict[Tuple[str, str], List[float]]) -> nx.MultiGraph:
        """
        Create a multi-graph from a dictionary of node pairs and associated edge weights.

        For example, creates a multi-graph where node edges represent animal relationship weights at different
        timepoints.

        :param Dict[Tuple[str, str], List[float]] data: A dictionary where keys are tuples representing node pairs, and values are lists of edge weights associated with each pair.
        :returns nx.MultiGraph: A NetworkX multigraph with nodes and edges specified by the input data. Each edge is labeled and weighted based on the provided information.

        :example:
        >>> data = {('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]}
        >>> NetworkMixin().create_multigraph(data=data)
        """

        check_instance(
            source=NetworkMixin.create_multigraph.__name__,
            instance=data,
            accepted_types=dict,
        )
        results, dict_data, G = {}, [], nx.MultiGraph()
        for k, v in data.items():
            check_instance(
                source=NetworkMixin.create_multigraph.__name__,
                instance=k,
                accepted_types=tuple,
            )
            check_instance(
                source=f"{NetworkMixin.create_multigraph.__name__} {v}",
                instance=v,
                accepted_types=list,
            )
            dict_data.append(len(v))
        check_iterable_length(
            source=f"{NetworkMixin.create_multigraph.__name__} data",
            val=len(list(set([x for x in dict_data]))),
            exact_accepted_length=1,
        )
        for node_names in data.keys():
            G.add_node(node_names[0])
            G.add_node(node_names[1])
        for node_names, edge_weights in data.items():
            for edge_cnt, edge_weight in enumerate(edge_weights):
                G.add_edge(
                    node_names[0],
                    node_names[1],
                    weight=edge_weight,
                    label=f"edge_{edge_cnt}",
                )
        return G

    @staticmethod
    def graph_page_rank(
        graph: nx.Graph,
        weights: Optional[str] = "weight",
        alpha: Optional[float] = 0.85,
        max_iter: Optional[int] = 100,
    ) -> Dict[str, float]:
        """
        Calculate the PageRank of nodes in a graph.

        :example:
        >>> graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
        >>> NetworkMixin().graph_page_rank(graph=graph)
        """

        check_instance(
            source=NetworkMixin.graph_page_rank.__name__,
            instance=graph,
            accepted_types=nx.Graph,
        )
        check_str(
            name=f"{NetworkMixin.graph_page_rank.__name__} weights", value=weights
        )
        check_float(name=f"{NetworkMixin.graph_page_rank.__name__} alpha", value=alpha)
        check_int(
            name=f"{NetworkMixin.graph_page_rank.__name__} max_iter",
            value=max_iter,
            min_value=1,
        )
        edge_weights = tuple(
            set(
                itertools.chain.from_iterable(
                    d.keys() for *_, d in graph.edges(data=True)
                )
            )
        )
        check_str(
            name=f"{NetworkMixin.graph_page_rank.__name__} weights",
            value=weights,
            options=edge_weights,
        )

        return nx.pagerank(graph, alpha=alpha, max_iter=max_iter, weight=weights)

    @staticmethod
    def graph_katz_centrality(
        graph: nx.Graph,
        weights: Optional[str] = "weight",
        alpha: Optional[float] = 0.85,
    ):
        """
        Katz centrality is an algorithm in NetworkX that measures the relative influence of a node in a network.

        See networkx documentation

        :example:
        >>> graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
        >>> NetworkMixin().graph_katz_centrality(graph=graph)
        """
        check_instance(
            source=NetworkMixin.graph_katz_centrality.__name__,
            instance=graph,
            accepted_types=nx.Graph,
        )
        check_str(
            name=f"{NetworkMixin.graph_katz_centrality.__name__} weights", value=weights
        )
        check_float(
            name=f"{NetworkMixin.graph_katz_centrality.__name__} alpha", value=alpha
        )
        edge_weights = tuple(
            set(
                itertools.chain.from_iterable(
                    d.keys() for *_, d in graph.edges(data=True)
                )
            )
        )
        check_str(
            name=f"{NetworkMixin.graph_katz_centrality.__name__} weights",
            value=weights,
            options=edge_weights,
        )

        return nx.katz_centrality_numpy(graph, alpha=alpha, weight=weights)

    @staticmethod
    def graph_current_flow_closeness_centrality(
        graph: nx.Graph, weights: Optional[str] = "weight"
    ):
        """

        :example:
        >>> graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
        >>> NetworkMixin().graph_current_flow_closeness_centrality(graph=graph)
        """

        check_instance(
            source=NetworkMixin.graph_current_flow_closeness_centrality.__name__,
            instance=graph,
            accepted_types=nx.Graph,
        )
        check_str(
            name=f"{NetworkMixin.graph_current_flow_closeness_centrality.__name__} weights",
            value=weights,
        )
        edge_weights = tuple(
            set(
                itertools.chain.from_iterable(
                    d.keys() for *_, d in graph.edges(data=True)
                )
            )
        )
        check_str(
            name=f"{NetworkMixin.graph_current_flow_closeness_centrality.__name__} weights",
            value=weights,
            options=edge_weights,
        )
        return nx.current_flow_closeness_centrality(
            graph, weight=weights, solver="full"
        )

    @staticmethod
    def girvan_newman(
        graph: nx.Graph,
        levels: Optional[int] = 1,
        most_valuable_edge: Optional[object] = None,
    ):
        """
        :example:
        >>> graph = NetworkMixin.create_graph({ ('Animal_1', 'Animal_2'): 0.0, ('Animal_1', 'Animal_3'): 0.0, ('Animal_1', 'Animal_4'): 0.0, ('Animal_1', 'Animal_5'): 0.0, ('Animal_2', 'Animal_3'): 1.0, ('Animal_2', 'Animal_4'): 1.0, ('Animal_2', 'Animal_5'): 1.0, ('Animal_3', 'Animal_4'): 1.0, ('Animal_3', 'Animal_5'): 1.0, ('Animal_4', 'Animal_5'): 1.0})
        >>> NetworkMixin().girvan_newman(graph=graph, levels = 1)
        >>> [({'Animal_1'}, {'Animal_2', 'Animal_3', 'Animal_4', 'Animal_5'})]
        """

        check_instance(
            source=NetworkMixin.multigraph_page_rank.__name__,
            instance=graph,
            accepted_types=nx.Graph,
        )
        if levels > G.number_of_nodes():
            raise CountError(
                msg=f"Number of nodes ({G.number_of_nodes()}) is less than numer of girvan newman levels ({levels}).",
                source=NetworkMixin.__class__.__name__,
            )

        communities = list(
            nx.algorithms.community.girvan_newman(
                graph, most_valuable_edge=most_valuable_edge
            )
        )
        return communities[:levels]

    @staticmethod
    def multigraph_page_rank(
        graph: nx.MultiGraph,
        weights: Optional[str] = "weight",
        alpha: Optional[float] = 0.85,
        max_iter: Optional[int] = 100,
    ) -> Dict[str, List[float]]:
        """
        Calculate multi-graph PageRank scores for each node in a MultiGraph.

        For example, each node-pair in a graph has N undirected edges representing the weighted relationship between the two nodes atobserved point in time.
        Calculates the page rank of each node at each observed time point.

        :param nx.MultiGraph graph: The input MultiGraph, created by ``NetworkMixin.create_multigraph()``.

        :example:
        >>> multigraph = NetworkMixin().create_multigraph(data={('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]})
        >>> NetworkMixin().multigraph_page_rank(graph=multigraph)
        >>> {'Animal_1': [0.06122524589028524, 0.06122524589028524, 0.06122524589028524, 0.32739635847890775], 'Animal_2': [0.06122524589028524, 0.40816213116457223, 0.06122524589028524, 0.442259400816002], 'Animal_3': [0.40816213116457223, 0.06122524589028524, 0.40816213116457223, 0.04545454545454547], 'Animal_4': [0.06122524589028524, 0.40816213116457223, 0.06122524589028524, 0.13943514979599955], 'Animal_5': [0.40816213116457223, 0.06122524589028524, 0.40816213116457223, 0.04545454545454547]}
        """

        check_instance(
            source=NetworkMixin.multigraph_page_rank.__name__,
            instance=graph,
            accepted_types=nx.MultiGraph,
        )
        edge_labels = list(set(data["label"] for _, _, data in graph.edges(data=True)))
        check_iterable_length(
            source=f"{NetworkMixin.multigraph_page_rank.__name__} edge_labels",
            val=len(edge_labels),
            min=1,
        )
        check_str(
            name=f"{NetworkMixin.graph_page_rank.__name__} weights", value=weights
        )
        check_float(name=f"{NetworkMixin.graph_page_rank.__name__} alpha", value=alpha)
        check_int(
            name=f"{NetworkMixin.graph_page_rank.__name__} max_iter",
            value=max_iter,
            min_value=1,
        )
        results = {x: [] for x in list(graph.nodes())}
        for edge_label in edge_labels:
            filtered_graph = nx.Graph(
                graph.edge_subgraph(
                    [
                        (u, v, k)
                        for u, v, k, data in graph.edges(keys=True, data=True)
                        if data.get("label") == edge_label
                    ]
                )
            )
            page_ranks = NetworkMixin().graph_page_rank(
                graph=filtered_graph, weights=weights, alpha=alpha, max_iter=max_iter
            )
            for k, v in page_ranks.items():
                results[k].append(v)

    @staticmethod
    def visualize(
        graph: Union[nx.Graph, nx.MultiGraph],
        save_path: Optional[Union[str, os.PathLike]] = None,
        node_size: Optional[Union[float, Dict[str, float]]] = 25.0,
        palette: Optional[Union[str, Dict[str, str]]] = "magma",
        img_size: Optional[Tuple[int, int]] = (500, 500),
    ) -> Union[None, Network]:
        """
        Visualizes a network graph using the vis.js library and saves the result as an HTML file.

        .. raw:: html
           :file: ../docs/_static/img/network_ex.html

        .. note::
           Multi-networks created by ``simba.mixins.network_mixin.create_multigraph`` can be a little messy to look at. Instead,
           creates seperate objects and files with single edges from each time-point.

        :param Union[nx.Graph, nx.MultiGraph] graph: The input graph to be visualized.
        :param Optional[Union[str, os.PathLike]] save_path: The path to save the HTML file. If None, the graph is not saved but returned. Default: None.
        :param Optional[Union[float, Dict[str, float]]] node_size: The size of nodes. Can be a single float or a dictionary mapping node names to their respective sizes. Default: 25.0.
        :param Optional[Union[str, Dict[str, str]]] palette: The color palette for nodes. Can be a single string representing a palette name or a dictionary mapping node names to their respective colors. Default; magma.
        :param Optional[Tuple[int, int]] img_size: The size of the resulting image in pixels, represented as (width, height). Default: 500x500.

        :example:
        >>> graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
        >>> graph_pg = NetworkMixin().graph_page_rank(graph=graph)
        """

        check_instance(
            source=NetworkMixin.visualize.__name__,
            instance=graph,
            accepted_types=(nx.MultiGraph, nx.Graph),
        )
        check_instance(
            source=NetworkMixin.visualize.__name__,
            instance=node_size,
            accepted_types=(int, float, dict),
        )
        multi_graph = False
        if graph.is_multigraph() and save_path is not None:
            if not os.path.isdir(save_path):
                os.makedirs(save_path)
            multi_graph = True
        check_instance(
            source=NetworkMixin.visualize.__name__,
            instance=img_size,
            accepted_types=tuple,
        )
        for i in img_size:
            check_int(
                name=f"{NetworkMixin.visualize.__name__} img_size",
                value=i,
                min_value=100,
            )
        if isinstance(node_size, dict):
            if sorted(graph.nodes) != sorted(list(node_size.keys())):
                raise InvalidInputError(
                    msg=f"node_size keys do not match graph node names: {sorted(graph.nodes)} != {sorted(list(node_size.keys()))}"
                )
            for v in node_size.values:
                check_float(
                    name=f"{NetworkMixin.visualize.__name__} node_size",
                    value=v,
                    min_value=0,
                )
        else:
            check_int(
                name=f"{NetworkMixin.visualize.__name__} node size",
                value=node_size,
                min_value=1,
            )
        if isinstance(palette, dict):
            if sorted(graph.nodes) != sorted(list(palette.keys())):
                raise InvalidInputError(
                    msg=f"palette keys do not match graph node names: {sorted(graph.nodes)} != {sorted(list(node_size.keys()))}"
                )
            for v in palette.values():
                check_valid_hex_color(color_hex=str(v))
        else:
            clrs = create_color_palette(
                pallete_name=palette, as_hex=True, increments=len(list(graph.nodes()))
            )

        if not multi_graph:
            network_graph = Network(f"{img_size[0]}px", f"{img_size[1]}px")
            network_graph.set_edge_smooth("dynamic")
            for node_cnt, node_name in enumerate(graph):
                if isinstance(node_size, dict):
                    node_node_size = node_size[node_name]
                else:
                    node_node_size = node_size
                if isinstance(palette, dict):
                    node_clr = palette[node_name]
                else:
                    node_clr = clrs[node_cnt]
                network_graph.add_node(
                    n_id=node_name, shape="dot", color=node_clr, size=node_node_size
                )
            for source, target, edge_attrs in graph.edges(data=True):
                network_graph.add_edge(source, target, value=edge_attrs["weight"])
            if save_path is not None:
                network_graph.save_graph(save_path)

            return network_graph

        else:
            edge_labels = list(
                set(data["label"] for _, _, data in graph.edges(data=True))
            )
            check_iterable_length(
                source=f"{NetworkMixin.multigraph_page_rank.__name__} edge_labels",
                val=len(edge_labels),
                min=1,
            )
            for edge_label in edge_labels:
                network_graph = Network(f"{img_size[0]}px", f"{img_size[1]}px")
                network_graph.set_edge_smooth("dynamic")
                network_graph.force_atlas_2based()
                graph_save_path = os.path.join(save_path, f"{edge_label}.html")
                filtered_graph = nx.Graph(
                    graph.edge_subgraph(
                        [
                            (u, v, k)
                            for u, v, k, data in graph.edges(keys=True, data=True)
                            if data.get("label") == edge_label
                        ]
                    )
                )
                for node_cnt, node_name in enumerate(filtered_graph):
                    if isinstance(node_size, dict):
                        node_node_size = node_size[node_name]
                    else:
                        node_node_size = node_size
                    if isinstance(palette, dict):
                        node_clr = palette[node_name]
                    else:
                        node_clr = palette[node_cnt]
                    network_graph.add_node(
                        n_id=node_name, shape="dot", color=node_clr, size=node_node_size
                    )
                for source, target, edge_attrs in filtered_graph.edges(data=True):
                    network_graph.add_edge(source, target, value=edge_attrs["weight"])
                if save_path is not None:
                    network_graph.save_graph(graph_save_path)


graph = NetworkMixin.create_graph(
    {
        ("Animal_1", "Animal_2"): 0.0,
        ("Animal_1", "Animal_3"): 0.0,
        ("Animal_1", "Animal_4"): 0.0,
        ("Animal_1", "Animal_5"): 0.1,
        ("Animal_2", "Animal_3"): 1.0,
        ("Animal_2", "Animal_4"): 1.0,
        ("Animal_2", "Animal_5"): 1.5,
        ("Animal_3", "Animal_4"): 1.0,
        ("Animal_3", "Animal_5"): 1.0,
        ("Animal_4", "Animal_5"): 1.0,
    }
)

G = nx.Graph(graph)

NetworkMixin().graph_current_flow_closeness_centrality(graph=graph)
#
#
# communities = list(nx.algorithms.community.(G, weight='weight'))
#
# partition = nx.algorithms.community.kernighan_lin_bisection(G, weight='weight')
# partition = nx.algorithms.community.greedy_modularity_communities(G, weight='weight')
# partition = list(nx.algorithms.community.girvan_newman(G))
#
# tuple(sorted(c) for c in next(partition))

# graph_data = {
#     ('Animal_1', 'Animal_2'): 0.0,
#     ('Animal_1', 'Animal_3'): 0.0,
#     ('Animal_1', 'Animal_4'): 0.0,
#     ('Animal_1', 'Animal_5'): 0.0,
#     ('Animal_2', 'Animal_3'): 1.5,  # Increase weight
#     ('Animal_2', 'Animal_4'): 1.0,  # Increase weight
#     ('Animal_2', 'Animal_5'): 1.5,  # Increase weight
#     ('Animal_3', 'Animal_4'): 2.9,  # Increase weight
#     ('Animal_3', 'Animal_5'): 2.0,  # Increase weight
#     ('Animal_4', 'Animal_5'): 1.0   # Decrease weight
# }

# result = list(NetworkMixin().asyn_lpa_communities(graph=graph, weight='weight'))
#
#
# for i in list(result):
#     print(i)
#


# graph_pg = NetworkMixin().graph_page_rank(graph=graph)
# graph_clrs = find_ranked_colors(data=graph_pg, palette='jet', as_hex=True)
# NetworkMixin().visualize(graph=graph,
#                          save_path='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html/graph_101.html',
#                          palette=graph_clrs)


# multigraph = NetworkMixin().create_multigraph(data={('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]})
#
#
# #graph_pg = NetworkMixin().graph_page_rank(graph=graph)
# graph_clrs = find_ranked_colors(data=graph_pg, palette='jet', as_hex=True)
#
# NetworkMixin().visualize(graph=graph,
#                          save_path='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html.html',
#                          palette=graph_clrs)


# multigraph = NetworkMixin().create_multigraph(data={('Animal_1', 'Animal_2'): [0, 0, 0, 6], ('Animal_1', 'Animal_3'): [0, 0, 0, 0], ('Animal_1', 'Animal_4'): [0, 0, 0, 0], ('Animal_1', 'Animal_5'): [0, 0, 0, 0], ('Animal_2', 'Animal_3'): [0, 0, 0, 0], ('Animal_2', 'Animal_4'): [5, 0, 0, 2], ('Animal_2', 'Animal_5'): [0, 0, 0, 0], ('Animal_3', 'Animal_4'): [0, 0, 0, 0], ('Animal_3', 'Animal_5'): [0, 2, 22, 0], ('Animal_4', 'Animal_5'): [0, 0, 0, 0]})
# NetworkMixin().visualize(graph=multigraph, save_path='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html')


# graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
# NetworkMixin().graph_page_rank(graph=graph)
# multigraph = NetworkMixin().create_multigraph(data=data)


# multigraph = NetworkMixin().create_multigraph(data=data)
#
# #multigraph.edges.data()
#
# NetworkMixin().visualize(graph=multigraph, save_dir='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html')


#
# for video_name, G in self.graphs.items():
#     if node_colors is not None:
#         clrs = create_single_color_lst(pallete_name='magma', increments=len(self.animal_names), as_hex=True)
#         node_colors = {k: v for k, v in sorted(node_colors.items(), key=lambda item: item[1], reverse=True)}
#         for node_cnt, node_name in enumerate(node_colors.keys()):
#             node_colors[node_name] = clrs[node_cnt]
#
#     network_graph = Network(style_attr['size'][0], style_attr['size'][1])
#     for node_name, node_attrs in G.nodes(data=True):
#         network_graph.add_node(node_name, color=node_colors[node_name])
#
#     for source, target, edge_attrs in G.edges(data=True):
#         edge_attrs['value'] = edge_attrs['weight']
#         network_graph.add_edge(str(source), str(target), **edge_attrs)
#
#     network_graph.save_graph('nx.html')
#


#
#
# data = {('Animal_1', 'Animal_2'): 0.0,
#         ('Animal_1', 'Animal_3'): 0.2,
#         ('Animal_1', 'Animal_4'): 0.3,
#         ('Animal_2', 'Animal_3'): 0.9,
#         ('Animal_2', 'Animal_4'): 1.0,
#         ('Animal_3', 'Animal_4'): 0.4}
# graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
# NetworkMixin().graph_current_flow_closeness_centrality(graph=graph)
#


#
#
# data = {('Animal_1', 'Animal_2'): [0, 0, 0, 6],
#         ('Animal_1', 'Animal_3'): [0, 0, 0, 0],
#         ('Animal_1', 'Animal_4'): [0, 0, 0, 0],
#         ('Animal_1', 'Animal_5'): [0, 0, 0, 0],
#         ('Animal_2', 'Animal_3'): [0, 0, 0, 0],
#         ('Animal_2', 'Animal_4'): [5, 0, 0, 2],
#         ('Animal_2', 'Animal_5'): [0, 0, 0, 0],
#         ('Animal_3', 'Animal_4'): [0, 0, 0, 0],
#         ('Animal_3', 'Animal_5'): [0, 2, 22, 0],
#         ('Animal_4', 'Animal_5'): [0, 0, 0, 0]}


# graph = NetworkMixin.create_graph(data={('Animal_1', 'Animal_2'): 1.0, ('Animal_1', 'Animal_3'): 0.2, ('Animal_2', 'Animal_3'): 0.5})
# #NetworkMixin().graph_page_rank(graph=graph)
# multigraph = NetworkMixin().create_multigraph(data=data)
#
# #multigraph.edges.data()
#
# NetworkMixin().visualize(graph=multigraph, save_dir='/Users/simon/Desktop/envs/troubleshooting/ARES_data/Termite Test/project/project_data/network_html')


# NetworkMixin().multigraph_page_rank(graph=multigraph)


# NetworkMixin.page_rank(graph=graph)
# NetworkMixin.katz_centrality(graph=graph)
# NetworkMixin.current_flow_closeness_centrality(graph=graph)



__author__ = "Simon Nilsson"

import os
from copy import deepcopy

import numpy as np

from simba.mixins.config_reader import ConfigReader
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import \
    check_all_file_names_are_represented_in_video_log
from simba.utils.data import plug_holes_shortest_bout
from simba.utils.enums import TagNames
from simba.utils.errors import NoFilesFoundError, ParametersFileError
from simba.utils.printing import SimbaTimer, log_event, stdout_success
from simba.utils.read_write import get_fn_ext, read_df, write_df
from simba.utils.warnings import SkippingFileWarning


class InferenceBatch(TrainModelMixin, ConfigReader):
    """
    Run classifier inference on all files with the ``project_folder/csv/features_extracted`` directory.
    Results are stored in the ``project_folder/csv/machine_results`` directory of the SimBA project.

    :param str config_path: path to SimBA project config file in Configparser format

    Example
    ----------
    >>> _ = InferenceBatch(config_path='MyConfigPath').run()
    """

    def __init__(self, config_path: str):

        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        log_event(
            logger_name=str(self.__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        if len(self.feature_file_paths) == 0:
            raise NoFilesFoundError(
                "Zero files found in the project_folder/csv/features_extracted directory. Create features before running classifier.",
                source=self.__class__.__name__,
            )
        print(
            f"Analyzing {len(self.feature_file_paths)} file(s) with {self.clf_cnt} classifier(s)"
        )
        self.timer = SimbaTimer(start=True)
        self.model_dict = self.get_model_info(
            config=self.config, model_cnt=self.clf_cnt
        )

    def run(self):
        check_all_file_names_are_represented_in_video_log(
            video_info_df=self.video_info_df, data_paths=self.feature_file_paths
        )
        for file_cnt, file_path in enumerate(self.feature_file_paths):
            video_timer = SimbaTimer(start=True)
            _, file_name, _ = get_fn_ext(file_path)
            print("Analyzing video {}...".format(file_name))
            file_save_path = os.path.join(
                self.machine_results_dir, file_name + "." + self.file_type
            )
            in_df = read_df(file_path, self.file_type)
            x_df = self.drop_bp_cords(df=in_df).astype("float32")
            self.check_df_dataset_integrity(
                df=x_df, logs_path=self.logs_path, file_name=file_name
            )
            _, px_per_mm, fps = self.read_video_info(
                video_name=file_name, raise_error=False
            )
            out_df = deepcopy(in_df)
            for m, m_hyp in self.model_dict.items():
                if not os.path.isfile(m_hyp["model_path"]):
                    raise NoFilesFoundError(
                        msg=f"{m_hyp['model_path']} is not a VALID model file path",
                        source=self.__class__.__name__,
                    )
                probability_column = "Probability_" + m_hyp["model_name"]
                clf = self.read_pickle(file_path=m_hyp["model_path"])
                out_df[probability_column] = self.clf_predict_proba(
                    clf=clf,
                    x_df=x_df,
                    data_path=file_path,
                    model_name=m_hyp["model_name"],
                )
                out_df[m_hyp["model_name"]] = np.where(
                    out_df[probability_column] > m_hyp["threshold"], 1, 0
                )
                out_df = plug_holes_shortest_bout(
                    data_df=out_df,
                    clf_name=m_hyp["model_name"],
                    fps=fps,
                    shortest_bout=m_hyp["minimum_bout_length"],
                )
            write_df(out_df, self.file_type, file_save_path)
            video_timer.stop_timer()
            print(
                f"Predictions created for {file_name} (elapsed time: {video_timer.elapsed_time_str}) ..."
            )
        self.timer.stop_timer()
        stdout_success(
            msg="Machine predictions complete. Files saved in project_folder/csv/machine_results directory",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# test = InferenceBatch(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.run()

# test = RunModel(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.run_models()



__author__ = "Simon Nilsson"

import itertools
import os
from collections import defaultdict
from copy import deepcopy
from datetime import datetime
from typing import List, Optional, Union

import cv2
import h5py
import numpy as np
import pandas as pd
import scipy.io as sio
from numba import jit, prange

from simba.utils.enums import ConfigKey
from simba.utils.errors import (CountError, IntegerError, InvalidInputError,
                                NoDataError, NoFilesFoundError)
from simba.utils.read_write import get_fn_ext
from simba.utils.warnings import FrameRangeWarning


class PoseImporterMixin(object):
    """
    Methods for importing pose-estimation data.
    """

    def __init__(self):
        self.datetime = datetime.now().strftime("%Y%m%d%H%M%S")
        pass

    def initialize_multi_animal_ui(
        self,
        animal_bp_dict: dict,
        video_info: dict,
        data_df: pd.DataFrame,
        video_path: str,
    ):

        self.video_info, self.data_df, self.frame_no, self.add_spacer = (
            video_info,
            data_df,
            0,
            2,
        )
        self.animal_bp_dict, self.cap = animal_bp_dict, cv2.VideoCapture(video_path)
        _, self.video_name, _ = get_fn_ext(video_path)
        self.get_video_scalers(video_info=video_info)

    def get_video_scalers(self, video_info: dict):
        self.scalers = {}
        space_scale, radius_scale, resolution_scale, font_scale = 40, 10, 1500, 1.2
        max_video_dimension = max(video_info["width"], video_info["height"])
        self.scalers["circle"] = int(
            radius_scale / (resolution_scale / max_video_dimension)
        )
        self.scalers["font"] = font_scale / (resolution_scale / max_video_dimension)
        self.scalers["space"] = int(
            space_scale / (resolution_scale / max_video_dimension)
        )

    def find_data_files(self, dir: Union[str, os.PathLike], extensions: List[str]):

        data_paths = []
        paths = [f for f in next(os.walk(dir))[2] if not f[0] == "."]
        paths = [os.path.join(dir, f) for f in paths]
        for extension in extensions:
            for path in paths:
                if path.endswith(extension):
                    data_paths.append(path)
        if len(data_paths) == 0:
            raise NoDataError(
                msg=f"No files with {extensions} extensions found in {dir}.",
                source=self.__class__.__name__,
            )

        return data_paths

    def link_video_paths_to_data_paths(
        self,
        data_paths: List[str],
        video_paths: List[str],
        str_splits: Optional[List[str]] = None,
        filename_cleaning_func: Optional[object] = None,
    ) -> dict:
        """
        Given a list of paths to video files and a separate list of paths to data files, create a dictionary
        pairing each video file to a datafile based on the file names of the video and data file.

        :param List[str] data_paths: List of full paths to data files, e.g., CSV or H5 files.
        :param List[str] video_paths: List of full paths to video files, e.g., MP4 or AVI files.
        :param Optional[List[str]] str_splits: Optional list of substrings that the data_paths would need to be split at in order to find a matching video name. E.g., ['dlc_resnet50'].
        :param Optional[object] filename_cleaning_func: Optional filename cleaning function that the data_paths filenames would have to pass through in order to find a matching video name. E.g., ``simba.utils.read_write.clean_sleap_filename(filepath)``.
        :returns dict: Dictionary with the data/file name as keys, and the video and data paths as values.
        """

        results, video_names = {}, []
        for video_path in video_paths:
            _, video_file_name, _ = get_fn_ext(video_path)
            video_names.append(video_file_name.lower())
        for data_path in data_paths:
            _, data_file_name, _ = get_fn_ext(data_path)
            data_file_names = [data_file_name.lower()]
            if str_splits:
                for split_str in str_splits:
                    data_file_names.append(data_file_name.lower().split(split_str)[0])
            data_file_names = list(set(data_file_names))
            if filename_cleaning_func is not None:
                data_file_names = [filename_cleaning_func(x) for x in data_file_names]
            video_idx = [i for i, x in enumerate(video_names) if x in data_file_names]
            if len(video_idx) == 0:
                raise NoFilesFoundError(
                    msg=f"SimBA could not locate a video file in your SimBA project for data file {data_file_name}",
                    source=self.__class__.__name__,
                )
            _, video_name, _ = get_fn_ext(video_paths[video_idx[0]])
            results[video_name] = {
                "DATA": data_path,
                "VIDEO": video_paths[video_idx[0]],
            }
        return results

    def get_x_y_loc_of_mouse_click(self, event, x, y, flags, param):
        if event == 7:
            self.click_loc = (x, y)
            self.id_cords[self.cnt] = {}
            self.id_cords[self.cnt]["cord"] = self.click_loc
            self.id_cords[self.cnt]["name"] = self.animal_name

    def insert_all_bodyparts_into_img(self, img: np.ndarray, body_parts: dict):
        for animal, bp_data in body_parts.items():
            for bp_cnt, bp_tuple in enumerate(bp_data):
                try:
                    cv2.circle(
                        img,
                        bp_tuple,
                        self.scalers["circle"],
                        self.animal_bp_dict[animal]["colors"][bp_cnt],
                        -1,
                        lineType=cv2.LINE_AA,
                    )
                except Exception as err:
                    if type(err) == OverflowError:
                        raise IntegerError(
                            f"SimBA encountered a pose-estimated body-part located at pixel position {str(bp_tuple)}. "
                            "This value is too large to be converted to an integer. "
                            "Please check your pose-estimation data to make sure that it is accurate.",
                            source=self.__class__.__name__,
                        )

    def insert_animal_names(self):
        for animal_cnt, animal_data in self.id_cords.items():
            cv2.putText(
                self.new_frame,
                animal_data["name"],
                animal_data["cord"],
                cv2.FONT_HERSHEY_SIMPLEX,
                self.scalers["font"],
                (255, 255, 255),
                3,
            )

    def multianimal_identification(self):
        cv2.destroyAllWindows()
        self.cap.set(1, self.frame_no)
        self.all_frame_data = self.data_df.loc[self.frame_no, :]
        cv2.namedWindow("Define animal IDs", cv2.WINDOW_NORMAL)
        _, self.img = self.cap.read()
        self.img_overlay, self.img_bp_cords = deepcopy(self.img), defaultdict(list)
        for animal_cnt, (animal_name, animal_bps) in enumerate(
            self.animal_bp_dict.items()
        ):
            self.img_bp_cords[animal_name] = []
            for x_name, y_name in zip(animal_bps["X_bps"], animal_bps["Y_bps"]):
                self.img_bp_cords[animal_name].append(
                    tuple(
                        self.data_df.loc[self.frame_no, [x_name, y_name]].values.astype(
                            int
                        )
                    )
                )
        self.insert_all_bodyparts_into_img(
            img=self.img_overlay, body_parts=self.img_bp_cords
        )
        side_img = np.ones(
            (int(self.video_info["height"] / 2), self.video_info["width"], 3)
        )
        cv2.putText(
            side_img,
            f"Current video: {self.video_name}",
            (10, self.scalers["space"]),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.scalers["font"],
            (255, 255, 255),
            3,
        )
        cv2.putText(
            side_img,
            "Can you assign identities based on the displayed frame ?",
            (10, int(self.scalers["space"] * (self.add_spacer * 2))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.scalers["font"],
            (255, 255, 255),
            2,
        )
        cv2.putText(
            side_img,
            'Press "x" to display new, random, frame',
            (10, int(self.scalers["space"] * (self.add_spacer * 3))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.scalers["font"],
            (255, 255, 0),
            3,
        )
        cv2.putText(
            side_img,
            'Press "c" to continue to start assigning identities using this frame',
            (10, int(self.scalers["space"] * (self.add_spacer * 4))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.scalers["font"],
            (0, 255, 0),
            2,
        )
        self.img_concat = np.uint8(np.concatenate((self.img_overlay, side_img), axis=0))
        cv2.imshow("Define animal IDs", self.img_concat)
        cv2.resizeWindow(
            "Define animal IDs", self.video_info["height"], self.video_info["width"]
        )
        keyboard_choice = False
        while not keyboard_choice:
            k = cv2.waitKey(20)
            if (k == ord("x")) or (k == ord("X")):
                if self.frame_no + 50 > len(self.data_df):
                    FrameRangeWarning(
                        msg=f"Cannot proceed to new frame: the frame {self.frame_no+50} does not exist in video {self.video_name}. The video {self.video_name} has {len(self.data_df)} frames."
                    )
                else:
                    cv2.destroyWindow("Define animal IDs")
                    cv2.waitKey(0)
                    self.frame_no += 50
                    self.multianimal_identification()
                    break
            elif (k == ord("c")) or (k == ord("C")):
                cv2.destroyWindow("Define animal IDs")
                cv2.waitKey(0)
                self.choose_animal_ui()
                break

    def choose_animal_ui(self):
        self.id_cords = {}
        for cnt, animal in enumerate(self.animal_bp_dict.keys()):
            self.animal_name, self.cnt = animal, cnt
            self.new_overlay = deepcopy(self.img_overlay)
            cv2.namedWindow("Define animal IDs", cv2.WINDOW_NORMAL)
            self.side_img = np.ones(
                (int(self.video_info["height"] / 2), self.video_info["width"], 3)
            )
            cv2.putText(
                self.side_img,
                "Double left mouse click on:",
                (10, self.scalers["space"]),
                cv2.FONT_HERSHEY_SIMPLEX,
                self.scalers["font"],
                (255, 255, 255),
                3,
            )
            cv2.putText(
                self.side_img,
                animal,
                (10, int(self.scalers["space"] * (self.add_spacer * 2))),
                cv2.FONT_HERSHEY_SIMPLEX,
                self.scalers["font"],
                (255, 255, 0),
                3,
            )
            for id in self.id_cords.keys():
                cv2.putText(
                    self.new_overlay,
                    self.id_cords[id]["name"],
                    self.id_cords[id]["cord"],
                    cv2.FONT_HERSHEY_SIMPLEX,
                    self.scalers["font"],
                    (255, 255, 255),
                    3,
                )
            self.new_overlay = np.uint8(
                np.concatenate((self.new_overlay, self.side_img), axis=0)
            )
            cv2.imshow("Define animal IDs", self.new_overlay)
            cv2.resizeWindow(
                "Define animal IDs", self.video_info["height"], self.video_info["width"]
            )
            while cnt not in self.id_cords.keys():
                cv2.setMouseCallback(
                    "Define animal IDs", self.get_x_y_loc_of_mouse_click
                )
                cv2.waitKey(200)
        self.confirm_ui()

    def confirm_ui(self):
        cv2.destroyAllWindows()
        cv2.namedWindow("Define animal IDs", cv2.WINDOW_NORMAL)
        cv2.resizeWindow(
            "Define animal IDs", self.video_info["height"], self.video_info["width"]
        )
        self.new_frame = deepcopy(self.img)
        self.side_img = np.ones(
            (int(self.video_info["height"] / 2), self.video_info["width"], 3)
        )
        cv2.putText(
            self.side_img,
            f"Current video: {self.video_name}",
            (10, self.scalers["space"]),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.scalers["font"],
            (255, 255, 255),
            3,
        )
        cv2.putText(
            self.side_img,
            "Are you happy with your assigned identities ?",
            (10, int(self.scalers["space"] * (self.add_spacer * 2))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.scalers["font"],
            (255, 255, 255),
            2,
        )
        cv2.putText(
            self.side_img,
            'Press "c" to continue (to finish, or proceed to the next video)',
            (10, int(self.scalers["space"] * (self.add_spacer * 3))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.scalers["font"],
            (255, 255, 0),
            2,
        )
        cv2.putText(
            self.side_img,
            'Press "x" to re-start assigning identities',
            (10, int(self.scalers["space"] * (self.add_spacer * 4))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.scalers["font"],
            (0, 255, 255),
            2,
        )
        self.insert_all_bodyparts_into_img(
            img=self.new_frame, body_parts=self.img_bp_cords
        )
        self.insert_animal_names()
        self.img_concat = np.uint8(
            np.concatenate((self.new_frame, self.side_img), axis=0)
        )
        cv2.imshow("Define animal IDs", self.img_concat)
        cv2.resizeWindow(
            "Define animal IDs", self.video_info["height"], self.video_info["width"]
        )
        keyboard_choice = False
        while not keyboard_choice:
            k = cv2.waitKey(20)
            if (k == ord("x")) or (k == ord("X")):
                if self.frame_no + 50 > len(self.data_df):
                    FrameRangeWarning(
                        msg=f"Cannot proceed to new frame: the frame {self.frame_no+50} does not exist in video {self.video_name}. The video {self.video_name} has {len(self.data_df)} frames."
                    )
                else:
                    cv2.destroyWindow("Define animal IDs")
                    cv2.waitKey(0)
                    self.frame_no += 50
                    self.multianimal_identification()
                    break
            elif (k == ord("c")) or (k == ord("C")):
                cv2.destroyAllWindows()
                cv2.waitKey(0)
                self.cap.release()
                self.find_closest_animals()
                break

    def find_closest_animals(self):
        self.animal_order = {}
        for animal_number, animal_click_data in self.id_cords.items():
            animal_name, animal_cord = (
                animal_click_data["name"],
                animal_click_data["cord"],
            )
            closest_animal = {}
            closest_animal["animal_name"] = None
            closest_animal["body_part_name"] = None
            closest_animal["distance"] = np.inf
            for other_animal_name, animal_bps in self.animal_bp_dict.items():
                animal_bp_names_x = self.animal_bp_dict[other_animal_name]["X_bps"]
                animal_bp_names_y = self.animal_bp_dict[other_animal_name]["Y_bps"]
                for x_col, y_col in zip(animal_bp_names_x, animal_bp_names_y):
                    bp_location = (
                        int(self.all_frame_data[x_col]),
                        int(self.all_frame_data[y_col]),
                    )
                    distance = np.sqrt(
                        (animal_cord[0] - bp_location[0]) ** 2
                        + (animal_cord[1] - bp_location[1]) ** 2
                    )
                    if distance < closest_animal["distance"]:
                        closest_animal["animal_name"] = other_animal_name
                        closest_animal["body_part_name"] = (x_col, y_col)
                        closest_animal["distance"] = distance
            self.animal_order[animal_number] = closest_animal
        self.check_intergity_of_chosen_animal_order()
        self.organize_results()
        self.reinsert_multi_idx_columns()

    def organize_results(self):
        self.out_df = pd.DataFrame()
        for animal_cnt, animal_data in self.animal_order.items():
            closest_animal_dict = self.animal_bp_dict[animal_data["animal_name"]]
            x_cols, y_cols, p_cols = (
                closest_animal_dict["X_bps"],
                closest_animal_dict["Y_bps"],
                closest_animal_dict["P_bps"],
            )
            for x_col, y_col, p_cols in zip(x_cols, y_cols, p_cols):
                df = self.data_df[[x_col, y_col, p_cols]]
                self.out_df = pd.concat([self.out_df, df], axis=1)

    def reinsert_multi_idx_columns(self):
        multi_idx_cols = []
        for col_idx in range(len(self.out_df.columns)):
            multi_idx_cols.append(
                tuple(("IMPORTED_POSE", "IMPORTED_POSE", self.out_df.columns[col_idx]))
            )
        self.out_df.columns = pd.MultiIndex.from_tuples(
            multi_idx_cols, names=("scorer", "bodypart", "coords")
        )

    def insert_multi_idx_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        multi_idx_cols = []
        for col_idx in range(len(df.columns)):
            multi_idx_cols.append(
                tuple(("IMPORTED_POSE", "IMPORTED_POSE", df.columns[col_idx]))
            )
        df.columns = pd.MultiIndex.from_tuples(
            multi_idx_cols, names=("scorer", "bodypart", "coords")
        )
        return df

    def check_intergity_of_chosen_animal_order(self):
        for click_key_combination in itertools.combinations(
            list(self.animal_order.keys()), 2
        ):
            click_n, click_n1 = click_key_combination[0], click_key_combination[1]
            animal_1, animal_2 = (
                self.animal_order[click_n]["animal_name"],
                self.animal_order[click_n1]["animal_name"],
            )
            if animal_1 == animal_2:
                raise InvalidInputError(
                    msg=f"The animal most proximal to click number {str(click_n)} is animal named {animal_1}. The animal most proximal to click number {str(click_n1)} is also animal {animal_2}."
                    "Please indicate which animal is which using a video frame where the animals are clearly separated",
                    source=self.__class__.__name__,
                )

    def intertwine_probability_cols(self, data: pd.DataFrame) -> pd.DataFrame:
        results = pd.DataFrame()
        for animal_name, animal_bps in self.animal_bp_dict.items():
            x_cols, y_cols, p_cols = (
                animal_bps["X_bps"],
                animal_bps["Y_bps"],
                animal_bps["P_bps"],
            )
            for x_col, y_col, p_col in zip(x_cols, y_cols, p_cols):
                df = data[[x_col, y_col, p_col]]
                results = pd.concat([results, df], axis=1)
        return results

    def __update_config_animal_cnt(self):
        self.config.set(
            ConfigKey.GENERAL_SETTINGS.value,
            ConfigKey.ANIMAL_CNT.value,
            str(self.animal_cnt),
        )
        with open(self.project_path, "w+") as f:
            self.config.write(f)
        f.close()

    def update_bp_headers_file(self):
        new_headers = []
        for animal_name in self.animal_bp_dict.keys():
            for bp in self.animal_bp_dict[animal_name]["X_bps"]:
                if animal_name not in bp:
                    new_headers.append("{}_{}".format(animal_name, bp[:-2]))
                else:
                    new_headers.append(bp[:-2])
        new_bp_df = pd.DataFrame(new_headers)
        new_bp_df.to_csv(self.body_parts_path, index=False, header=False)

    def update_config_animal_names(self):
        self.config.set(
            ConfigKey.MULTI_ANIMAL_ID_SETTING.value,
            ConfigKey.MULTI_ANIMAL_IDS.value,
            ",".join(str(x) for x in self.id_lst),
        )
        with open(self.project_path, "w+") as f:
            self.config.write(f)
        f.close()

    @staticmethod
    @jit(nopython=True)
    def transpose_multi_animal_table(
        data: np.array, idx: np.array, animal_cnt: int
    ) -> np.array:
        results = np.full((np.max(idx[:, 1]), data.shape[1] * animal_cnt), 0.0)
        for i in prange(np.max(idx[:, 1])):
            for j in prange(animal_cnt):
                data_idx = np.argwhere((idx[:, 0] == j) & (idx[:, 1] == i)).flatten()
                if len(data_idx) == 1:
                    animal_frm_data = data[data_idx[0]]
                else:
                    animal_frm_data = np.full((data.shape[1]), 0.0)
                results[i][
                    j * animal_frm_data.shape[0] : j * animal_frm_data.shape[0]
                    + animal_frm_data.shape[0]
                ] = animal_frm_data
        return results

    def read_apt_trk_file(self, file_path: str):
        print("Reading data using scipy.io...")
        try:
            trk_coordinates = sio.loadmat(file_path)["pTrk"]
            track_cnt = trk_coordinates.shape[3]
            data = [trk_coordinates[..., i] for i in range(track_cnt)]

        except NotImplementedError:
            print("Failed to read data using scipy.io. Reading data using h5py...")
            with h5py.File(file_path, "r") as trk_dict:
                trk_list = list(trk_dict["pTrk"])
                t_second = np.array(trk_list)
                if len(t_second.shape) > 3:
                    t_third = np.swapaxes(t_second, 0, 3)
                    trk_coordinates = np.swapaxes(t_third, 1, 2)
                    track_cnt = trk_coordinates.shape[3]
                    data = [trk_coordinates[..., i] for i in range(track_cnt)]
                else:
                    data = np.swapaxes(t_second, 0, 2)
                    track_cnt = 1
        if track_cnt != self.animal_cnt:
            raise CountError(
                msg=f"There are {str(track_cnt)} tracks in the .trk file {file_path}. But your SimBA project expects {str(self.animal_cnt)} tracks.",
                source=self.__class__.__name__,
            )

        if self.animal_cnt != 1:
            animal_df_lst = []
            for animal in data:
                m, n, r = animal.shape
                out_arr = np.column_stack(
                    (np.repeat(np.arange(m), n), animal.reshape(m * n, -1))
                )
                animal_df_lst.append(
                    pd.DataFrame(out_arr).T.iloc[1:].reset_index(drop=True)
                )
                self.data_df = pd.concat(animal_df_lst, axis=1).fillna(0).astype(np.int)
        else:
            m, n, r = data.shape
            out_arr = np.column_stack(
                (np.repeat(np.arange(m), n), data.reshape(m * n, -1))
            )
            self.data_df = (
                pd.DataFrame(out_arr).T.iloc[1:].reset_index(drop=True).astype(np.int)
            )

        p_cols = pd.DataFrame(
            1, index=self.data_df.index, columns=self.data_df.columns[1::2] + 0.5
        )
        return pd.concat([self.data_df, p_cols], axis=1)  # .sort_index(axis=1)

    # def import_log(self):
    #     self.import_log = pd.DataFrame(columns=['VIDEO', 'IMPORT_TIME', 'IMPORT_SOURCE', 'INTERPOLATION_SETTING', 'SMOOTHING_SETTING'])
    #

    # import_log.loc[len(import_log)] = [self.file_name, 'MADLC', str(self.interpolation_settings), str(self.smoothing_settings)]



import glob
import os
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser

import numpy as np
import pandas as pd
from numba import jit, prange

from simba.drop_bp_cords import getBpNames
from simba.read_config_unit_tests import (check_file_exist_and_readable,
                                          check_if_filepath_list_is_empty,
                                          read_config_entry, read_config_file)
from simba.rw_dfs import read_df

COL_NAMES = [
    "Low_prob_detections_0.1",
    "Low_prob_detections_0.5",
    "Low_prob_detections_0.75",
]


def add_probability_features_to_annotated_data(config_path: str):
    @jit(nopython=True)
    def count_values_in_range(data: np.array, ranges: np.array):
        results = np.full((data.shape[0], ranges.shape[0]), 0)
        for i in prange(data.shape[0]):
            for j in prange(ranges.shape[0]):
                lower_bound, upper_bound = ranges[j][0], ranges[j][1]
                results[i][j] = data[i][
                    np.logical_and(data[i] >= lower_bound, data[i] <= upper_bound)
                ].shape[0]
        return results

    check_file_exist_and_readable(file_path=config_path)
    config = read_config_file(ini_path=config_path)
    project_path = read_config_entry(
        config, "General settings", "project_path", data_type="folder_path"
    )
    targets_dir = os.path.join(project_path, "csv", "targets_inserted")
    files_found = glob.glob(targets_dir + "/*.csv")
    check_if_filepath_list_is_empty(
        filepaths=files_found, error_msg=f"ERROR: No files found in {targets_dir}"
    )
    x_cols, y_cols, pcols = getBpNames(config_path)

    for file_path in files_found:
        df = read_df(file_path=file_path, file_type="csv").reset_index(drop=True)
        missing_cols = list(set(COL_NAMES) - set(df.columns))
        if len(missing_cols) == 0:
            print(f"SIMBA ERROR: {COL_NAMES} already exist in {file_path}.")
            continue
        results = pd.DataFrame(
            count_values_in_range(
                data=df.filter(pcols).values,
                ranges=np.array([[0.0, 0.1], [0.000000000, 0.5], [0.000000000, 0.75]]),
            ),
            columns=[
                "Low_prob_detections_0.1",
                "Low_prob_detections_0.5",
                "Low_prob_detections_0.75",
            ],
        )
        df = pd.concat([df, results], axis=1)
        df.to_csv(file_path)
        print(f"New features added to {file_path}...")

    print(
        f"COMPLETE: {str(len(COL_NAMES))} new features added to {len(files_found)} in the {targets_dir} directory"
    )


if __name__ == "__main__":
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-dir", "--directory", help="path to SimBA project config file")
    args = vars(parser.parse_args())
    add_probability_features_to_annotated_data(config_path=args["directory"])



__author__ = "Simon Nilsson"

import functools
import multiprocessing
import os
import platform
from collections import deque
from copy import deepcopy
from typing import Dict, List, Optional

import cv2
import numpy as np
import pandas as pd
from numba import jit, prange

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.data import find_frame_numbers_from_time_stamp
from simba.utils.errors import FrameRangeError, NoSpecifiedOutputError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import (concatenate_videos_in_folder, get_fn_ext,
                                    read_df, remove_a_folder)


class PathPlotterMulticore(ConfigReader, PlottingMixin):
    """
    Class for creating "path plots" videos and/or images detailing the movement paths of
    individual animals in SimBA. Uses multiprocessing.

    :param str config_path: Path to SimBA project config file in Configparser format
    :param bool frame_setting: If True, individual frames will be created.
    :param bool video_setting: If True, compressed videos will be created.
    :param bool last_frame: If True, png of the last frame will be created.
    :param List[str] files_found: Data paths to create path plots for (e.g., ['project_folder/csv/machine_results/MyVideo.csv'])
    :param dict animal_attr: Animal body-parts to use when creating paths and their colors.
    :param Optional[dict] input_style_attr: Plot sttributes (line thickness, color, etc..). If None, then autocomputed. Max lines will be set to 2s.
    :param Optional[dict] input_clf_attr: Dict reprenting classified behavior locations, their color and size. If None, then no classified behavior locations are shown.
    :param Optional[dict] slicing: If Dict, start time and end time of video slice to create path plot from. E.g., {'start_time': '00:00:01', 'end_time': '00:00:03'}. If None, creates path plot for entire video.
    :param int cores: Number of cores to use.

    .. note::
       `Visualization tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-11-visualizations>`__.

    .. important::
       See :meth:`simba.utils.lookups.get_color_dict` for keys for named color options.

    .. image:: _static/img/path_plot.png
       :width: 300
       :align: center

    :example:
    >>> input_style_attr = {'width': 'As input', 'height': 'As input', 'line width': 5, 'font size': 5, 'font thickness': 2, 'circle size': 5, 'bg color': 'White', 'max lines': 100}
    >>> animal_attr = {0: ['Ear_right_1', 'Red']}
    >>> input_clf_attr = {0: ['Attack', 'Black', 'Size: 30'], 1: ['Sniffing', 'Red', 'Size: 30']}
    >>> path_plotter = PathPlotterMulticore(config_path=r'MyConfigPath', frame_setting=False, video_setting=True, style_attr=style_attr, animal_attr=animal_attr, files_found=['project_folder/csv/machine_results/MyVideo.csv'], cores=5, clf_attr=clf_attr)
    >>> path_plotter.run()
    """

    def __init__(
        self,
        config_path: str,
        frame_setting: bool,
        video_setting: bool,
        last_frame: bool,
        files_found: List[str],
        cores: int,
        input_style_attr: Optional[Dict] = None,
        animal_attr: Dict[int, List[str]] = None,
        input_clf_attr: Optional[Dict[int, List[str]]] = None,
        slicing: Optional[Dict] = None,
    ):
        if platform.system() == "Darwin":
            multiprocessing.set_start_method("spawn", force=True)

        if (not frame_setting) and (not video_setting) and (not last_frame):
            raise NoSpecifiedOutputError(
                msg="SIMBA ERROR: Please choice to create path frames and/or video path plots"
            )
        ConfigReader.__init__(self, config_path=config_path)
        PlottingMixin.__init__(self)
        (
            self.video_setting,
            self.frame_setting,
            self.input_style_attr,
            self.files_found,
            self.animal_attr,
            self.input_clf_attr,
            self.last_frame,
            self.cores,
        ) = (
            video_setting,
            frame_setting,
            input_style_attr,
            files_found,
            animal_attr,
            input_clf_attr,
            last_frame,
            cores,
        )
        self.no_animals_path_plot, self.clf_attr, self.slicing = (
            len(animal_attr.keys()),
            None,
            slicing,
        )
        if not os.path.exists(self.path_plot_dir):
            os.makedirs(self.path_plot_dir)
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/machine_results directory. To plot paths without performing machine classifications, use path plotter functions in [ROI] tab.",
        )
        print(f"Processing {str(len(self.files_found))} videos...")

    def run(self):
        """
        Method to create path plot videos and/or frames.Results are store in the
        'project_folder/frames/path_plots' directory of the SimBA project.
        """

        for file_cnt, file_path in enumerate(self.files_found):
            video_timer = SimbaTimer(start=True)
            _, self.video_name, _ = get_fn_ext(file_path)
            self.video_info, _, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.data_df = read_df(file_path, self.file_type)
            if self.slicing:
                frm_numbers = find_frame_numbers_from_time_stamp(
                    start_time=self.slicing["start_time"],
                    end_time=self.slicing["end_time"],
                    fps=self.fps,
                )
                if len(set(frm_numbers) - set(self.data_df.index)) > 0:
                    raise FrameRangeError(
                        msg=f'The chosen time-period ({self.slicing["start_time"]} - {self.slicing["end_time"]}) does not exist in {self.video_name}.'
                    )
                else:
                    self.data_df = self.data_df.loc[
                        frm_numbers[0] : frm_numbers[-1]
                    ].reset_index(drop=True)
            self.__get_styles()
            self.temp_folder = os.path.join(self.path_plot_dir, self.video_name, "temp")
            self.save_frame_folder_dir = os.path.join(
                self.path_plot_dir, self.video_name
            )
            if self.frame_setting:
                if os.path.exists(self.save_frame_folder_dir):
                    remove_a_folder(self.save_frame_folder_dir)
                if not os.path.exists(self.save_frame_folder_dir):
                    os.makedirs(self.save_frame_folder_dir)
            if self.video_setting:
                self.video_folder = os.path.join(self.path_plot_dir, self.video_name)
                if os.path.exists(self.temp_folder):
                    remove_a_folder(self.temp_folder)
                    remove_a_folder(self.video_folder)
                os.makedirs(self.temp_folder)
                self.save_video_path = os.path.join(
                    self.path_plot_dir, self.video_name + ".mp4"
                )

            if self.input_clf_attr:
                clf_names, self.clf_attr = [], {}
                self.clf_attr["attr"] = deepcopy(self.input_clf_attr)
                for v in self.input_clf_attr.values():
                    clf_names.append(v[0])
                self.clf_attr["data"] = self.data_df[clf_names]
                self.clf_attr["positions"] = self.data_df[
                    [self.animal_attr[0][0] + "_x", self.animal_attr[0][0] + "_y"]
                ]

            if self.last_frame:
                self.__get_deque_lookups()
                _ = self.make_path_plot(
                    data_df=self.data_df,
                    video_info=self.video_info,
                    style_attr=self.style_attr,
                    deque_dict=self.deque_dict,
                    clf_attr=self.clf_attr,
                    save_path=os.path.join(
                        self.path_plot_dir, self.video_name + "_final_frame.png"
                    ),
                )

            if self.video_setting or self.frame_setting:
                data_arr = np.array(list(self.data_df.index)).reshape(-1, 1)
                for animal_cnt, animal_data in self.animal_attr.items():
                    bp_x_name = "{}_{}".format(animal_data[0], "x")
                    bp_y_name = "{}_{}".format(animal_data[0], "y")
                    data_arr = np.hstack(
                        (
                            data_arr,
                            self.data_df[[bp_x_name, bp_y_name]].astype(int).values,
                        )
                    )
                    if animal_cnt == 0 and self.clf_attr:
                        self.clf_attr["positions"] = deepcopy(data_arr[:, 1:3])
                data_arr = self.__split_array_into_max_lines(
                    data=data_arr, max_lines=self.style_attr["max lines"]
                )
                data_arr = np.array_split(data_arr, self.cores)
                data = []
                for cnt, i in enumerate(data_arr):
                    data.append(self.__insert_group_idx_column(data=i, group=cnt))
                frm_per_core = data[0].shape[0]

                print(
                    "Creating path plots, multiprocessing (chunksize: {}, cores: {})...".format(
                        str(self.multiprocess_chunksize), str(self.cores)
                    )
                )
                with multiprocessing.Pool(
                    self.cores, maxtasksperchild=self.maxtasksperchild
                ) as pool:
                    constants = functools.partial(
                        self.path_plot_mp,
                        video_setting=self.video_setting,
                        video_name=self.video_name,
                        frame_setting=self.frame_setting,
                        video_save_dir=self.temp_folder,
                        frame_folder_dir=self.save_frame_folder_dir,
                        style_attr=self.style_attr,
                        fps=self.fps,
                        animal_attr=self.animal_attr,
                        video_info=self.video_info,
                        clf_attr=self.clf_attr,
                    )
                    for cnt, result in enumerate(
                        pool.imap(
                            constants, data, chunksize=self.multiprocess_chunksize
                        )
                    ):
                        print(
                            "Image {}/{}, Video {}/{}...".format(
                                str(int(frm_per_core * (result + 1))),
                                str(len(self.data_df)),
                                str(file_cnt + 1),
                                str(len(self.files_found)),
                            )
                        )

                    pool.terminate()
                    pool.join()

                if self.video_setting:
                    print("Joining {} multiprocessed video...".format(self.video_name))
                    concatenate_videos_in_folder(
                        in_folder=self.temp_folder, save_path=self.save_video_path
                    )

                video_timer.stop_timer()
                print(
                    "Path plot video {} complete (elapsed time: {}s) ...".format(
                        self.video_name, video_timer.elapsed_time_str
                    )
                )

        self.timer.stop_timer()
        stdout_success(
            msg=f"Path plot visualizations for {str(len(self.files_found))} videos created in project_folder/frames/output/path_plots directory",
            elapsed_time=self.timer.elapsed_time_str,
        )

    @staticmethod
    @jit(nopython=True)
    def __split_array_into_max_lines(data: np.array, max_lines: int):
        results = np.full((data.shape[0], max_lines, data.shape[1]), np.nan, data.dtype)
        for i in prange(data.shape[0]):
            start = int(i - max_lines)
            if start < 0:
                start = 0
            frm_data = data[start:i, :]

            missing_cnt = max_lines - frm_data.shape[0]

            if missing_cnt > 0:
                frm_data = np.vstack(
                    (
                        np.full((missing_cnt, frm_data.shape[1]), -1.0, frm_data.dtype),
                        frm_data,
                    )
                )

            results[i] = frm_data
        return results

    @staticmethod
    @jit(nopython=True)
    def __insert_group_idx_column(data: np.array, group: int):
        results = np.full((data.shape[0], data.shape[1], data.shape[2] + 1), np.nan)
        group_col = np.full((data.shape[1], 1), group)
        for frm_idx in prange(data.shape[0]):
            results[frm_idx] = np.hstack((group_col, data[frm_idx]))
        return results

    def __get_styles(self):
        self.style_attr = {}
        if self.input_style_attr is not None:
            self.style_attr["bg color"] = self.color_dict[
                self.input_style_attr["bg color"]
            ]
            if self.input_style_attr["max lines"] == "entire video":
                self.style_attr["max lines"] = len(self.data_df)
            else:
                self.style_attr["max lines"] = int(
                    int(self.input_style_attr["max lines"])
                    * (int(self.video_info["fps"].values[0]) / 1000)
                )
            self.style_attr["font thickness"] = self.input_style_attr["font thickness"]
            self.style_attr["line width"] = self.input_style_attr["line width"]
            self.style_attr["font size"] = self.input_style_attr["font size"]
            self.style_attr["circle size"] = self.input_style_attr["circle size"]
            if self.input_style_attr["width"] == "As input":
                self.style_attr["width"], self.style_attr["height"] = int(
                    self.video_info["Resolution_width"].values[0]
                ), int(self.video_info["Resolution_height"].values[0])
            else:
                pass
        else:
            space_scaler, radius_scaler, res_scaler, font_scaler = 25, 10, 1500, 0.8
            self.style_attr["width"] = int(
                self.video_info["Resolution_width"].values[0]
            )
            self.style_attr["height"] = int(
                self.video_info["Resolution_height"].values[0]
            )
            max_res = max(self.style_attr["width"], self.style_attr["height"])
            self.style_attr["circle size"] = int(radius_scaler / (res_scaler / max_res))
            self.style_attr["font size"] = font_scaler / (res_scaler / max_res)
            self.style_attr["bg color"] = self.color_dict["White"]
            self.style_attr["max lines"] = len(self.data_df)
            self.style_attr["font thickness"] = 2
            self.style_attr["line width"] = 2

        self.style_attr["animal names"] = []
        self.style_attr["animal clrs"] = []
        for animal_cnt, animal_data in self.animal_attr.items():
            self.style_attr["animal names"].append(
                self.find_animal_name_from_body_part_name(
                    bp_name=animal_data[0], bp_dict=self.animal_bp_dict
                )
            )

        for animal_cnt, animal_data in self.animal_attr.items():
            self.style_attr["animal clrs"].append(
                self.color_dict[self.animal_attr[animal_cnt][1]]
            )

    def __get_deque_lookups(self):
        self.deque_dict = {}
        for animal_cnt, animal in enumerate(self.style_attr["animal names"]):
            self.deque_dict[animal] = {}
            self.deque_dict[animal]["deque"] = deque(
                maxlen=self.style_attr["max lines"]
            )
            self.deque_dict[animal]["bp"] = self.animal_attr[animal_cnt][0]
            self.deque_dict[animal]["clr"] = self.color_dict[
                self.animal_attr[animal_cnt][1]
            ]


# style_attr = {'width': 'As input',
#               'height': 'As input',
#               'line width': 5, 'font size': 5, 'font thickness': 2, 'circle size': 5, 'bg color': 'White', 'max lines': 'entire video'}
# animal_attr = {0: ['Ear_right_1', 'Red'], 1: ['Ear_right_2', 'Green']}
# clf_attr = {0: ['Attack', 'Black', 'Size: 30']}
# # #
# # # clf_attr = None
# # # style_attr = None
# #
# path_plotter = PathPlotterMulticore(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                     frame_setting=False,
#                                     video_setting=True,
#                                     last_frame=True,
#                                     input_clf_attr=clf_attr,
#                                     input_style_attr=style_attr,
#                                     animal_attr=animal_attr,
#                                     files_found=['/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/csv/machine_results/Together_1.csv'],
#                                     cores=5,
#                                     slicing = {'start_time': '00:00:01', 'end_time': '00:00:03'})
# path_plotter.run()


# style_attr = None
# #style_attr = {'width': 'As input', 'height': 'As input', 'line width': 5, 'font size': 5, 'font thickness': 2, 'circle size': 5, 'bg color': 'White', 'max lines': 10000}
# animal_attr = {0: ['mouth', 'Red']}
# clf_attr = {0: ['Freezing', 'Black', 'Size: 10'], 1: ['Normal Swimming', 'Red', 'Size: 10']}
# #
# # clf_attr = None
#
#
# path_plotter = PathPlotterMulticore(config_path='/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
#                                     frame_setting=False,
#                                     video_setting=True,
#                                     last_frame=True,
#                                     input_clf_attr=clf_attr,
#                                     input_style_attr=style_attr,
#                                     animal_attr=animal_attr,
#                                     files_found=['/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/machine_results/SF8.csv'],
#                                     cores=5)
# path_plotter.run()



import copy
import glob
import os
import time
from tkinter import *

import cv2
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.roi_tools.ROI_image import ROI_image_class
from simba.roi_tools.ROI_move_shape import move_edge, update_all_tags
from simba.roi_tools.ROI_multiply import create_emty_df
from simba.roi_tools.ROI_size_calculations import (circle_size_calc,
                                                   polygon_size_calc,
                                                   rectangle_size_calc)
from simba.ui.tkinter_functions import hxtScrollbar
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext
from simba.utils.warnings import NoDataFoundWarning


class ROI_definitions(ConfigReader):
    """
    Launch ROI user-interface for drawing user-defined shapes in a video.

    Parameters
    ----------
    config_path: str
        path to SimBA project config file in Configparser format
    video_path: str
        path to video file for which ROIs should be defined.

    Notes
    ----------
    `ROI tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/ROI_tutorial_new.md>`__.

    Examples
    ----------
    >>> _ = ROI_definitions(config_path='MyProjectConfig', video_path='MyVideoPath')

    """

    def __init__(self, config_path: str, video_path: str, image_data: ROI_image_class):
        ConfigReader.__init__(self, config_path=config_path)
        self.video_path = video_path
        _, self.file_name, self.file_ext = get_fn_ext(self.video_path)
        self.roi_data_folder = os.path.join(self.logs_path, "measures")
        if not os.path.exists(self.roi_data_folder):
            os.makedirs(self.roi_data_folder)
        self.other_video_paths = glob.glob(self.video_dir + "/*.mp4") + glob.glob(
            self.video_dir + "/*.avi"
        )
        self.other_video_paths.remove(video_path)
        self.other_video_file_names = []
        for video in self.other_video_paths:
            self.other_video_file_names.append(os.path.basename(video))
        self.master_win_h, self.master_win_w = 1080, 750
        self.video_info, self.curr_px_mm, self.curr_fps = self.read_video_info(
            video_name=self.file_name
        )
        self.root = Tk()
        self.root.minsize(self.master_win_w, self.master_win_h)
        self.screen_width = self.root.winfo_screenwidth()
        self.screen_height = self.root.winfo_screenheight()
        self.default_top_left_x = self.screen_width - self.master_win_w
        self.root.geometry(
            "%dx%d+%d+%d"
            % (self.master_win_w, self.master_win_h, self.default_top_left_x, 0)
        )
        self.root.wm_title("Region of Interest Settings")
        self.root.protocol("WM_DELETE_WINDOW", self.on_close)
        self.shape_thickness_list = list(range(1, 26))
        self.ear_tag_size_list = list(range(1, 26))
        self.select_color = "red"
        self.non_select_color = "black"
        self.video_ROIs = ["None"]
        self.c_shape = None
        self.stored_interact = None
        self.stored_shape = None
        self.img_no = 1
        self.image_data = image_data

        self.named_shape_colors = self.image_data.colors
        self.window_menus()
        self.root.lift()
        self.master = Canvas(hxtScrollbar(self.root))
        self.master.pack(fill="both", expand=True)
        self.show_video_info()
        self.select_img()
        self.apply_from_other_videos_menu()
        self.select_shape()
        self.select_shape_attr()
        self.select_shape_name()
        self.interact_menus()
        self.draw_menu()
        self.save_menu()
        self.cap = cv2.VideoCapture(self.video_path)
        if not self.cap.isOpened():
            print("cant open video")
            return
        self.image_data.update_working_frame(self.cap, True, self.file_name)
        self.video_frame_count = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))
        self.get_all_ROI_names()
        if len(self.video_ROIs) > 0:
            self.update_delete_ROI_menu()

        self.master.mainloop()

    def on_close(self):
        self.cap.release()
        self.image_data.reset()
        self.master.destroy()
        self.root.destroy()

    def show_video_info(self):
        self.video_info_frame = LabelFrame(
            self.master,
            text="Video information",
            font=("Arial", 14, "bold"),
            padx=5,
            pady=5,
        )
        self.video_info_frame.grid_configure(ipadx=55)
        self.video_name_lbl_1 = Label(
            self.video_info_frame, text="Video name: ", font=("Arial", 10)
        ).grid(row=0, column=0)
        self.video_name_lbl_2 = Label(
            self.video_info_frame, text=str(self.file_name), font=("Arial", 10, "bold")
        )

        self.video_ext_lbl_1 = Label(
            self.video_info_frame, text="Video format: ", font=("Arial", 10)
        ).grid(row=0, column=2)
        self.video_ext_lbl_2 = Label(
            self.video_info_frame, text=str(self.file_ext), font=("Arial", 10, "bold")
        )

        self.video_fps_lbl_1 = Label(
            self.video_info_frame, text="FPS: ", font=("Arial", 10)
        ).grid(row=0, column=4)
        self.video_fps_lbl_2 = Label(
            self.video_info_frame, text=str(self.curr_fps), font=("Arial", 10, "bold")
        )

        self.video_frame_lbl_1 = Label(
            self.video_info_frame, text="Display frame #: ", font=("Arial", 10)
        ).grid(row=0, column=6)
        self.video_frame_lbl_2 = Label(
            self.video_info_frame, text=str(self.img_no), font=("Arial", 10, "bold")
        )

        self.video_frame_time_1 = Label(
            self.video_info_frame, text="Display frame (s): ", font=("Arial", 10)
        ).grid(row=0, column=8)
        self.video_frame_time_2 = Label(
            self.video_info_frame,
            text=str(round((self.img_no / self.curr_fps), 2)),
            font=("Arial", 10, "bold"),
        )

        self.video_info_frame.grid(row=0, sticky=W)
        self.video_name_lbl_2.grid(row=0, column=1)
        self.video_ext_lbl_2.grid(row=0, column=3)
        self.video_fps_lbl_2.grid(row=0, column=5)
        self.video_frame_lbl_2.grid(row=0, column=7)
        self.video_frame_time_2.grid(row=0, column=9)

    def select_img(self):
        self.img_no_frame = LabelFrame(
            self.master, text="Change image", font=("Arial", 16, "bold"), padx=5, pady=5
        )
        self.img_no_frame.grid_configure(ipadx=100)
        self.pos_1s = Button(
            self.img_no_frame,
            text="+1s",
            fg=self.non_select_color,
            command=lambda: self.set_current_image("plus"),
        )
        self.neg_1s = Button(
            self.img_no_frame,
            text="-1s",
            fg=self.non_select_color,
            command=lambda: self.set_current_image("minus"),
        )
        self.reset_btn = Button(
            self.img_no_frame,
            text="Reset first frame",
            fg=self.non_select_color,
            command=lambda: self.set_current_image("reset"),
        )
        self.seconds_fw_label = Label(self.img_no_frame, text="Seconds forward: ")
        self.seconds_fw_entry = Entry(self.img_no_frame, width=4)
        self.custom_run_seconds = Button(
            self.img_no_frame,
            text="Move",
            fg=self.non_select_color,
            command=lambda: self.set_current_image("custom"),
        )
        self.img_no_frame.grid(row=1, sticky=W)
        self.pos_1s.grid(row=1, column=0, sticky=W, pady=10, padx=10)
        self.neg_1s.grid(row=1, column=1, sticky=W, pady=10, padx=10)
        self.seconds_fw_label.grid(row=1, column=2, sticky=W, pady=10)
        self.seconds_fw_entry.grid(row=1, column=3, sticky=W, pady=10)
        self.custom_run_seconds.grid(row=1, column=4, sticky=W, pady=10)
        self.reset_btn.grid(row=1, column=5, sticky=W, pady=10, padx=10)

    def set_current_image(self, stride):
        if stride == "plus":
            img_no = self.img_no + self.curr_fps
            if (img_no > 0) and (img_no < self.video_frame_count):
                self.img_no = img_no
                self.pos_1s.configure(fg=self.select_color)
                self.neg_1s.configure(fg=self.non_select_color)
                self.custom_run_seconds.configure(fg=self.non_select_color)

        if stride == "minus":
            img_no = self.img_no - self.curr_fps
            if (img_no > 0) and (img_no < self.video_frame_count):
                self.img_no = img_no
                self.pos_1s.configure(fg=self.non_select_color)
                self.neg_1s.configure(fg=self.select_color)
                self.custom_run_seconds.configure(fg=self.non_select_color)

        if stride == "reset":
            self.img_no = 1

        if stride == "custom":
            img_no = self.img_no + int(self.curr_fps * int(self.seconds_fw_entry.get()))
            if (img_no > 0) and (img_no < self.video_frame_count):
                self.img_no = img_no
                self.pos_1s.configure(fg=self.non_select_color)
                self.neg_1s.configure(fg=self.non_select_color)
                self.custom_run_seconds.configure(fg=self.select_color)

        self.video_frame_lbl_2.config(text=str(self.img_no))
        self.video_frame_time_2.config(
            text=str(round((self.img_no / self.curr_fps), 2))
        )
        self.image_data.update_frame_no(self.img_no)

    def get_other_videos_w_data(self):
        self.other_videos_w_ROIs = []
        if os.path.isfile(self.roi_coordinates_path):
            for shape_type in ["rectangles", "circleDf", "polygons"]:
                c_df = pd.read_hdf(self.roi_coordinates_path, key=shape_type)
                if len(c_df) > 0:
                    self.other_videos_w_ROIs = list(
                        set(self.other_videos_w_ROIs + list(c_df["Video"].unique()))
                    )
        if len(self.other_videos_w_ROIs) == 0:
            self.other_videos_w_ROIs = ["None"]

    def get_all_ROI_names(self):
        self.video_ROIs = []
        for shape in [
            self.image_data.out_rectangles,
            self.image_data.out_circles,
            self.image_data.out_polygon,
        ]:
            for e in shape:
                shape_type = e["Shape_type"]
                shape_name = e["Name"]
                self.video_ROIs.append(shape_type + ": " + shape_name)

    def apply_rois_from_other_video(self):
        target_video = self.selected_other_video.get()
        if target_video != "None":
            if os.path.isfile(self.roi_coordinates_path):
                for shape_type in ["rectangles", "circleDf", "polygons"]:
                    c_df = pd.read_hdf(self.roi_coordinates_path, key=shape_type)
                    if len(c_df) > 0:
                        c_df = c_df[c_df["Video"] == target_video].reset_index(
                            drop=True
                        )
                        c_df["Video"] = self.file_name
                        c_df = c_df.to_dict("records")
                        if shape_type == "rectangles":
                            for r in c_df:
                                self.image_data.out_rectangles.append(r)
                        if shape_type == "circleDf":
                            for c in c_df:
                                self.image_data.out_circles.append(c)
                        if shape_type == "polygons":
                            for p in c_df:
                                self.image_data.out_polygon.append(p)
                self.get_all_ROI_names()
                self.update_delete_ROI_menu()
                self.image_data.insert_all_ROIs_into_image()

    def apply_from_other_videos_menu(self):
        self.get_other_videos_w_data()
        self.apply_from_other_video = LabelFrame(
            self.master,
            text="Apply shapes from another video",
            font=("Arial", 16, "bold"),
            padx=5,
            pady=5,
        )
        self.select_video_label = Label(
            self.apply_from_other_video, text="Select video: "
        ).grid(row=1, column=0)
        self.selected_other_video = StringVar()
        self.selected_other_video.set(self.other_videos_w_ROIs[0])
        self.video_dropdown = OptionMenu(
            self.apply_from_other_video,
            self.selected_other_video,
            *self.other_videos_w_ROIs,
        )
        self.apply_button = Button(
            self.apply_from_other_video,
            text="Apply",
            fg=self.non_select_color,
            command=lambda: self.apply_rois_from_other_video(),
        )
        self.apply_from_other_video.grid(row=7, sticky=W)
        self.video_dropdown.config(direction=RIGHT)
        self.video_dropdown.grid(row=1, column=1, sticky="ew", pady=20)
        self.apply_button.grid(row=1, column=3, sticky=W, pady=10)

    def select_shape(self):
        self.new_shape_frame = LabelFrame(
            self.master,
            text="New shape",
            font=("Arial", 16, "bold"),
            padx=5,
            pady=5,
            bd=5,
        )
        self.shape_frame = LabelFrame(
            self.new_shape_frame,
            text="Shape type",
            font=("Arial", 14, "bold"),
            padx=5,
            pady=5,
        )
        self.rectangle_button = Button(
            self.shape_frame,
            text="Rectangle",
            fg=self.non_select_color,
            command=lambda: self.set_current_shape("rectangle"),
        )
        self.circle_button = Button(
            self.shape_frame,
            text="Circle",
            fg=self.non_select_color,
            command=lambda: self.set_current_shape("circle"),
        )
        self.polygon_button = Button(
            self.shape_frame,
            text="Polygon",
            fg=self.non_select_color,
            command=lambda: self.set_current_shape("polygon"),
        )

        self.new_shape_frame.grid(row=3, sticky=W)
        self.shape_frame.grid(row=1, sticky=W)
        self.rectangle_button.grid(row=1, sticky=W, pady=10, padx=10)
        self.circle_button.grid(row=1, column=1, sticky=W, pady=10, padx=10)
        self.polygon_button.grid(row=1, column=2, sticky=W, pady=10, padx=10)

    def select_shape_attr(self):
        self.shape_attr_frame = LabelFrame(
            self.new_shape_frame,
            text="Shape attributes",
            font=("Arial", 16, "bold"),
            padx=5,
            pady=5,
        )
        self.shape_attr_frame.grid_configure(ipadx=50)
        self.thickness_label = Label(self.shape_attr_frame, text="Shape thickness: ")
        self.color_label = Label(self.shape_attr_frame, text="Shape color: ")
        self.shape_thickness = IntVar()
        self.shape_thickness.set(5)
        self.shape_thickness_dropdown = OptionMenu(
            self.shape_attr_frame,
            self.shape_thickness,
            *self.shape_thickness_list,
            command=None,
        )
        self.shape_thickness_dropdown.config(width=3)

        self.ear_tag_sizes_lbl = Label(self.shape_attr_frame, text="Ear tag size: ")
        self.ear_tag_size = IntVar()
        self.ear_tag_size.set(10)
        self.ear_tag_size_dropdown = OptionMenu(
            self.shape_attr_frame, self.ear_tag_size, *list(self.ear_tag_size_list)
        )

        self.color_var = StringVar()
        self.color_var.set("Red")
        self.color_dropdown = OptionMenu(
            self.shape_attr_frame, self.color_var, *list(self.named_shape_colors.keys())
        )

        self.shape_attr_frame.grid(row=2, sticky=W, pady=10)
        self.thickness_label.grid(row=1, column=0)
        self.shape_thickness_dropdown.grid(
            row=1, column=1, sticky=W, pady=10, padx=(0, 10)
        )
        self.ear_tag_sizes_lbl.grid(row=1, column=2)
        self.ear_tag_size_dropdown.grid(
            row=1, column=3, sticky=W, pady=10, padx=(0, 10)
        )
        self.color_label.grid(row=1, column=4)
        self.color_dropdown.grid(row=1, column=5, sticky=W, pady=10)

    def select_shape_name(self):
        self.set_shape_name = LabelFrame(
            self.new_shape_frame,
            text="Shape name",
            font=("Arial", 16, "bold"),
            padx=5,
            pady=5,
        )
        self.set_shape_name.grid_configure(ipadx=105)
        self.name_label = Label(self.set_shape_name, text="Shape name: ").grid(
            row=1, column=0
        )
        self.name_box = Entry(self.set_shape_name, width=55)
        self.set_shape_name.grid(row=3, sticky=W, pady=10)
        self.name_box.grid(row=1, column=2, sticky=W, pady=10)

    def interact_menus(self):
        self.interact_frame = LabelFrame(
            self.master,
            text="Shape interaction",
            font=("Arial", 16, "bold"),
            padx=5,
            pady=5,
        )
        self.interact_frame.grid_configure(ipadx=30)
        self.move_shape_button = Button(
            self.interact_frame,
            text="Move shape",
            fg=self.non_select_color,
            command=lambda: self.set_interact_state("move_shape"),
        )
        self.zoom_in_button = Button(
            self.interact_frame,
            text="Zoom IN",
            fg=self.non_select_color,
            state=DISABLED,
            command=lambda: self.set_interact_state("zoom_in"),
        )
        self.zoom_out_button = Button(
            self.interact_frame,
            text="Zoom OUT",
            fg=self.non_select_color,
            state=DISABLED,
            command=lambda: self.set_interact_state("zoom_out"),
        )
        self.zoom_home = Button(
            self.interact_frame,
            text="Zoom HOME",
            fg=self.non_select_color,
            state=DISABLED,
            command=lambda: self.set_interact_state("zoom_home"),
        )
        self.zoom_pct_label = Label(self.interact_frame, text="Zoom %: ").grid(
            row=1, column=5, padx=(10, 0)
        )
        self.zoom_pct = Entry(self.interact_frame, width=4, state=DISABLED)
        self.zoom_pct.insert(0, 10)
        self.pan = Button(
            self.interact_frame,
            text="Pan",
            fg=self.non_select_color,
            state=DISABLED,
            command=lambda: self.set_interact_state("pan"),
        )
        self.shape_info_btn = Button(
            self.interact_frame,
            text="Show shape info.",
            fg=self.non_select_color,
            command=lambda: self.show_shape_information(),
        )

        self.interact_frame.grid(row=6, sticky=W)
        self.move_shape_button.grid(row=1, column=0, sticky=W, pady=10, padx=10)
        self.pan.grid(row=1, column=1, sticky=W, pady=10, padx=10)
        self.zoom_in_button.grid(row=1, column=2, sticky=W, pady=10, padx=10)
        self.zoom_out_button.grid(row=1, column=3, sticky=W, pady=10, padx=10)
        self.zoom_home.grid(row=1, column=4, sticky=W, pady=10, padx=10)
        self.zoom_pct.grid(row=1, column=6, sticky=W, pady=10)
        self.shape_info_btn.grid(row=1, column=7, sticky=W, pady=10)

    def call_remove_ROI(self):
        self.shape_info_btn.configure(text="Show shape info.")
        self.apply_delete_button.configure(fg=self.select_color)
        self.image_data.remove_ROI(self.selected_video.get())
        self.video_ROIs.remove(self.selected_video.get())
        if len(self.video_ROIs) == 0:
            self.video_ROIs = ["None"]
        self.selected_video.set(self.video_ROIs[0])
        self.update_delete_ROI_menu()

    def draw_menu(self):
        self.draw_frame = LabelFrame(
            self.master, text="Draw", font=("Arial", 16, "bold"), padx=5, pady=5
        )
        self.draw_button = Button(
            self.draw_frame,
            text="Draw",
            fg=self.non_select_color,
            command=lambda: self.create_draw(),
        )
        self.delete_all_rois_btn = Button(
            self.draw_frame,
            text="Delete ALL",
            fg=self.non_select_color,
            command=lambda: self.call_delete_all_rois(),
        )
        self.select_roi_label = Label(self.draw_frame, text="Select ROI: ")
        self.selected_video = StringVar()
        self.selected_video.set(self.video_ROIs[0])
        self.roi_dropdown = OptionMenu(
            self.draw_frame, self.selected_video, *self.video_ROIs
        )
        self.apply_delete_button = Button(
            self.draw_frame,
            text="Delete ROI",
            fg=self.non_select_color,
            command=lambda: self.call_remove_ROI(),
        )
        self.duplicate_ROI_btn = Button(
            self.draw_frame,
            text="Duplicate ROI",
            fg=self.non_select_color,
            command=lambda: self.call_duplicate_ROI(),
        )
        self.chg_attr_btn = Button(
            self.draw_frame,
            text="Change ROI",
            fg=self.non_select_color,
            command=lambda: self.ChangeAttrMenu(self, self.image_data),
        )

        self.draw_frame.grid(row=5, sticky=W)
        self.draw_button.grid(row=1, column=1, sticky=W, pady=2, padx=10)
        self.delete_all_rois_btn.grid(row=1, column=2, sticky=W, pady=2, padx=10)
        self.select_roi_label.grid(row=1, column=3, sticky=W, pady=2, padx=(10, 0))
        self.roi_dropdown.grid(row=1, column=4, sticky=W, pady=2, padx=(0, 10))
        self.apply_delete_button.grid(row=1, column=5, sticky=W, pady=2, padx=10)
        self.duplicate_ROI_btn.grid(row=1, column=6, sticky=W, pady=2, padx=10)
        self.chg_attr_btn.grid(row=1, column=7, sticky=W, pady=2, padx=10)

    def show_shape_information(self):
        if (
            len(self.image_data.out_rectangles)
            + len(self.image_data.out_circles)
            + len(self.image_data.out_polygon)
            == 0
        ):
            print("No shapes to print info for.")

        elif self.shape_info_btn.cget("text") == "Show shape info.":
            if len(self.image_data.out_rectangles) > 0:
                self.rectangle_size_dict = {}
                self.rectangle_size_dict["Rectangles"] = {}
                for rectangle in self.image_data.out_rectangles:
                    self.rectangle_size_dict["Rectangles"][rectangle["Name"]] = (
                        rectangle_size_calc(rectangle, self.curr_px_mm)
                    )
                self.image_data.rectangle_size_dict = self.rectangle_size_dict

            if len(self.image_data.out_circles) > 0:
                self.circle_size_dict = {}
                self.circle_size_dict["Circles"] = {}
                for circle in self.image_data.out_circles:
                    self.circle_size_dict["Circles"][circle["Name"]] = circle_size_calc(
                        circle, self.curr_px_mm
                    )
                self.image_data.circle_size_dict = self.circle_size_dict

            if len(self.image_data.out_polygon) > 0:
                self.polygon_size_dict = {}
                self.polygon_size_dict["Polygons"] = {}
                for polygon in self.image_data.out_polygon:
                    self.polygon_size_dict["Polygons"][polygon["Name"]] = (
                        polygon_size_calc(polygon, self.curr_px_mm)
                    )
                self.image_data.polygon_size_dict = self.polygon_size_dict

            self.image_data.insert_all_ROIs_into_image(show_size_info=True)
            self.shape_info_btn.configure(text="Hide shape info.")

        elif self.shape_info_btn.cget("text") == "Hide shape info.":
            self.shape_info_btn.configure(text="Show shape info.")
            self.image_data.insert_all_ROIs_into_image()

    def save_menu(self):
        self.save_frame = LabelFrame(
            self.master, text="Save", font=("Arial", 16, "bold"), padx=5, pady=5
        )
        self.save_button = Button(
            self.save_frame,
            text="Save ROI data",
            fg=self.non_select_color,
            command=lambda: self.save_data(),
        )
        self.save_frame.grid(row=8, sticky=W)
        self.save_button.grid(row=1, column=3, sticky=W, pady=10)

    def set_current_shape(self, c_shape):
        self.c_shape = c_shape
        self.shape_info_btn.configure(text="Show shape info.")
        if self.c_shape == self.stored_shape:
            self.rectangle_button.configure(fg=self.non_select_color)
            self.circle_button.configure(fg=self.non_select_color)
            self.polygon_button.configure(fg=self.non_select_color)
            self.stored_shape = None
        else:
            if c_shape == "rectangle":
                self.rectangle_button.configure(fg=self.select_color)
                self.circle_button.configure(fg=self.non_select_color)
                self.polygon_button.configure(fg=self.non_select_color)

            if c_shape == "circle":
                self.rectangle_button.configure(fg=self.non_select_color)
                self.circle_button.configure(fg=self.select_color)
                self.polygon_button.configure(fg=self.non_select_color)

            if c_shape == "polygon":
                self.rectangle_button.configure(fg=self.non_select_color)
                self.circle_button.configure(fg=self.non_select_color)
                self.polygon_button.configure(fg=self.select_color)
            self.stored_shape = c_shape

    def reset_selected_buttons(self, category):
        if category == "interact" or "all":
            self.move_shape_button.configure(fg=self.non_select_color)
            self.zoom_in_button.configure(fg=self.non_select_color)
            self.zoom_out_button.configure(fg=self.non_select_color)
            self.pan.configure(fg=self.non_select_color)
            self.zoom_home.configure(fg=self.non_select_color)
            self.stored_interact = None

    def set_interact_state(self, c_interact):
        self.shape_info_btn.configure(text="Show shape info.")
        if c_interact == self.stored_interact:
            self.move_shape_button.configure(fg=self.non_select_color)
            self.zoom_in_button.configure(fg=self.non_select_color)
            self.zoom_out_button.configure(fg=self.non_select_color)
            self.pan.configure(fg=self.non_select_color)
            self.zoom_home.configure(fg=self.non_select_color)
            self.stored_interact = None

        else:
            if c_interact == "move_shape":
                if self.image_data.no_shapes > 0:
                    self.move_shape_button.configure(fg=self.select_color)
                    self.zoom_in_button.configure(fg=self.non_select_color)
                    self.zoom_out_button.configure(fg=self.non_select_color)
                    self.zoom_home.configure(fg=self.non_select_color)
                    self.pan.configure(fg=self.non_select_color)
                else:
                    self.reset_selected_buttons("interact")
                    c_interact = None
                    NoDataFoundWarning(msg="You have no shapes that can be moved.")

            if c_interact == "zoom_in":
                self.move_shape_button.configure(fg=self.non_select_color)
                self.zoom_in_button.configure(fg=self.select_color)
                self.zoom_out_button.configure(fg=self.non_select_color)
                self.zoom_home.configure(fg=self.non_select_color)
                self.pan.configure(fg=self.non_select_color)

            if c_interact == "zoom_out":
                self.move_shape_button.configure(fg=self.non_select_color)
                self.zoom_in_button.configure(fg=self.non_select_color)
                self.zoom_out_button.configure(fg=self.select_color)
                self.zoom_home.configure(fg=self.non_select_color)
                self.pan.configure(fg=self.non_select_color)

            if c_interact == "zoom_home":
                self.move_shape_button.configure(fg=self.non_select_color)
                self.zoom_in_button.configure(fg=self.non_select_color)
                self.zoom_out_button.configure(fg=self.non_select_color)
                self.zoom_home.configure(fg=self.select_color)
                self.pan.configure(fg=self.non_select_color)

            if c_interact == "pan":
                self.move_shape_button.configure(fg=self.non_select_color)
                self.zoom_in_button.configure(fg=self.non_select_color)
                self.zoom_out_button.configure(fg=self.non_select_color)
                self.zoom_home.configure(fg=self.non_select_color)
                self.pan.configure(fg=self.select_color)
            self.stored_interact = c_interact
        self.image_data.interact_functions(self.stored_interact, zoom_val=0)
        self.reset_selected_buttons("interact")

    def call_delete_all_rois(self):
        self.shape_info_btn.configure(text="Show shape info.")
        if (
            len(self.image_data.out_rectangles)
            + len(self.image_data.out_circles)
            + len(self.image_data.out_polygon)
            == 0
        ):
            NoDataFoundWarning(msg="SimBA finds no ROIs to delete.")
        else:
            self.image_data.out_rectangles = []
            self.image_data.out_circles = []
            self.image_data.out_polygon = []
            self.video_ROIs = ["None"]
            self.selected_video.set(self.video_ROIs[0])
            self.update_delete_ROI_menu()
            self.image_data.insert_all_ROIs_into_image()

    def get_duplicate_shape_name(self):
        c_no = 1
        while True:
            self.new_name = self.current_shape_data["Name"] + "_copy_" + str(c_no)
            if str(self.shape_type) + ": " + self.new_name in self.video_ROIs:
                c_no += 1
            else:
                self.new_shape_data["Name"] = (
                    str(self.shape_type) + ": " + self.new_name
                )
                break

    def get_duplicate_coords(self):
        if self.shape_type == "Rectangle":
            self.new_shape_x = int(
                self.current_shape_data["topLeftX"] + self.duplicate_jump_size
            )
            self.new_shape_y = int(
                self.current_shape_data["topLeftY"] + self.duplicate_jump_size
            )
            for shape in self.image_data.out_rectangles:
                if (shape["topLeftX"] == self.new_shape_x) and (
                    shape["topLeftY"] == self.new_shape_y
                ):
                    self.new_shape_x += self.duplicate_jump_size
                    self.new_shape_y += self.duplicate_jump_size
        if self.shape_type == "Circle":
            self.new_shape_x = int(
                self.current_shape_data["centerX"] + self.duplicate_jump_size
            )
            self.new_shape_y = int(
                self.current_shape_data["centerY"] + self.duplicate_jump_size
            )
            for shape in self.image_data.out_circles:
                if (shape["centerY"] == self.new_shape_x) and (
                    shape["centerY"] == self.new_shape_y
                ):
                    self.new_shape_x += self.duplicate_jump_size
                    self.new_shape_y += self.duplicate_jump_size
        if self.shape_type == "Polygon":
            self.new_shape_x = int(
                self.current_shape_data["centerX"] + self.duplicate_jump_size
            )
            self.new_shape_y = int(
                self.current_shape_data["centerY"] + self.duplicate_jump_size
            )
            for shape in self.image_data.out_polygon:
                if (shape["Center_X"] == self.new_shape_x) and (
                    shape["centerY"] == self.new_shape_y
                ):
                    self.new_shape_x += self.duplicate_jump_size
                    self.new_shape_y += self.duplicate_jump_size

    def call_duplicate_ROI(self):
        shape_name = self.selected_video.get().split(": ")
        self.shape_info_btn.configure(text="Show shape info.")
        if shape_name[0] != "None":
            all_roi_list = (
                self.image_data.out_rectangles
                + self.image_data.out_circles
                + self.image_data.out_polygon
            )
            self.shape_type, shape_name = shape_name[0], shape_name[1]
            self.current_shape_data = [
                d for d in all_roi_list if d.get("Name") == shape_name
            ][0]
            self.new_shape_data = copy.deepcopy(self.current_shape_data)
            self.get_duplicate_shape_name()
            self.get_duplicate_coords()
            if self.shape_type == "Rectangle":
                self.new_shape_data["topLeftX"] = self.new_shape_x
                self.new_shape_data["topLeftY"] = self.new_shape_y
                self.new_shape_data["Name"] = self.new_shape_data["Name"].split(
                    "Rectangle: ", 1
                )[-1]
                update_all_tags(self.new_shape_data)
                self.image_data.out_rectangles.append(self.new_shape_data)

            if self.shape_type == "Circle":
                self.new_shape_data["centerX"] = self.new_shape_x
                self.new_shape_data["centerY"] = self.new_shape_y
                self.new_shape_data["Name"] = self.new_shape_data["Name"].split(
                    "Circle: ", 1
                )[-1]
                update_all_tags(self.new_shape_data)
                self.image_data.out_circles.append(self.new_shape_data)

            if self.shape_type == "Polygon":
                move_edge(
                    self.new_shape_data,
                    "Center_tag",
                    (self.new_shape_x, self.new_shape_y),
                )
                self.new_shape_data["Name"] = self.new_shape_data["Name"].split(
                    "Polygon: ", 1
                )[-1]
                self.image_data.out_polygon.append(self.new_shape_data)

            self.video_ROIs.append(self.shape_type + ": " + self.new_shape_data["Name"])
            self.image_data.insert_all_ROIs_into_image()

            self.update_delete_ROI_menu()
        else:
            print("No ROI selected.")

    def create_draw(self):
        self.shape_info_btn.configure(text="Show shape info.")
        if self.stored_shape is None:
            raise TypeError("No shape type selected.")
        if not self.name_box.get():
            raise TypeError("No shape name selected.")
        if not self.name_box.get().strip():
            raise TypeError("Shape name contains only spaces.")

        c_draw_settings = {
            "Video_name": self.file_name,
            "Shape_type": self.stored_shape,
            "Name": self.name_box.get(),
            "Shape_thickness": self.shape_thickness.get(),
            "Shape_ear_tag_size": self.ear_tag_size.get(),
            "Shape_color_name": self.color_var.get(),
            "Shape_color_BGR": self.named_shape_colors[self.color_var.get()],
        }

        self.video_ROIs = self.image_data.initiate_draw(c_draw_settings)
        # cv2.setWindowProperty('Define shape', cv2.WND_PROP_TOPMOST, 1)

        self.update_delete_ROI_menu()

    def update_delete_ROI_menu(self):
        self.selected_video.set(self.video_ROIs[0])
        self.roi_dropdown = OptionMenu(
            self.draw_frame, self.selected_video, *self.video_ROIs
        )
        self.roi_dropdown.grid(row=1, column=4, sticky=W, pady=10)

    def save_data(self):
        if os.path.isfile(self.roi_coordinates_path):
            rectangles_found = pd.read_hdf(self.roi_coordinates_path, key="rectangles")
            circles_found = pd.read_hdf(self.roi_coordinates_path, key="circleDf")
            polygons_found = pd.read_hdf(self.roi_coordinates_path, key="polygons")
            other_vid_rectangles = rectangles_found[
                rectangles_found["Video"] != self.file_name
            ]
            other_vid_circles = circles_found[circles_found["Video"] != self.file_name]
            other_vid_polygons = polygons_found[
                polygons_found["Video"] != self.file_name
            ]

            new_rectangles = pd.DataFrame.from_dict(self.image_data.out_rectangles)
            new_circles = pd.DataFrame.from_dict(self.image_data.out_circles)
            new_polygons = pd.DataFrame.from_dict(self.image_data.out_polygon)

            if len(new_rectangles) > 0:
                out_rectangles = (
                    pd.concat([other_vid_rectangles, new_rectangles], axis=0)
                    .sort_values(by=["Video"])
                    .reset_index(drop=True)
                )
            else:
                out_rectangles = other_vid_rectangles.sort_values(
                    by=["Video"]
                ).reset_index(drop=True)

            if len(new_circles) > 0:
                out_circles = (
                    pd.concat([other_vid_circles, new_circles], axis=0)
                    .sort_values(by=["Video"])
                    .reset_index(drop=True)
                )
            else:
                out_circles = other_vid_circles.sort_values(by=["Video"]).reset_index(
                    drop=True
                )

            if len(new_polygons) > 0:
                out_polygons = (
                    pd.concat([other_vid_polygons, new_polygons], axis=0)
                    .sort_values(by=["Video"])
                    .reset_index(drop=True)
                )
            else:
                out_polygons = other_vid_polygons.sort_values(by=["Video"]).reset_index(
                    drop=True
                )

        else:
            out_rectangles = pd.DataFrame.from_dict(self.image_data.out_rectangles)
            out_circles = pd.DataFrame.from_dict(self.image_data.out_circles)
            out_polygons = pd.DataFrame.from_dict(self.image_data.out_polygon)

            if len(out_rectangles) == 0:
                out_rectangles = create_emty_df("rectangles")
            if len(out_circles) == 0:
                out_circles = create_emty_df("circleDf")
            if len(out_polygons) == 0:
                out_polygons = create_emty_df("polygons")

        store = pd.HDFStore(self.roi_coordinates_path, mode="w")
        store["rectangles"] = out_rectangles
        store["circleDf"] = out_circles
        store["polygons"] = out_polygons
        store.close()
        stdout_success(msg=f"ROI definitions saved for video: {self.file_name}")

    class ChangeAttrMenu:
        def __init__(self, shape_data, image_data):
            shape_name = shape_data.selected_video.get().split(": ")
            if shape_name[0] != "None":
                self.all_roi_list = (
                    shape_data.image_data.out_rectangles
                    + shape_data.image_data.out_circles
                    + shape_data.image_data.out_polygon
                )
                self.shape_type, self.shape_name = shape_name[0], shape_name[1]
                current_shape_data = [
                    d for d in self.all_roi_list if d.get("Name") == self.shape_name
                ][0]
                self.attr_win = Toplevel()
                self.attr_win.minsize(400, 300)
                self.attr_win.wm_title("Selected Shape Attributes")
                attr_lbl_frame = LabelFrame(
                    self.attr_win,
                    text="Attributes",
                    font=("Arial", 16, "bold"),
                    pady=5,
                    padx=5,
                    fg="black",
                )
                selected_shape_name_lbl = Label(self.attr_win, text="Shape name: ")
                self.selected_shape_name_entry_txt = StringVar()
                self.selected_shape_name_entry_txt.set(current_shape_data["Name"])

                selected_shape_name_entry = Entry(
                    self.attr_win,
                    width=25,
                    textvariable=self.selected_shape_name_entry_txt,
                )
                selected_shape_thickness_lbl = Label(
                    self.attr_win, text="Shape thickness: "
                )
                self.selected_shape_thickness = IntVar()
                self.selected_shape_thickness.set(current_shape_data["Thickness"])
                selected_shape_thickness_dropdown = OptionMenu(
                    self.attr_win,
                    self.selected_shape_thickness,
                    *list(shape_data.shape_thickness_list),
                )

                selected_shape_eartag_size_lbl = Label(
                    self.attr_win, text="Ear tag size: "
                )
                self.selected_shape_eartag_size = IntVar()
                self.selected_shape_eartag_size.set(current_shape_data["Ear_tag_size"])
                selected_shape_eartag_size_dropdown = OptionMenu(
                    self.attr_win,
                    self.selected_shape_eartag_size,
                    *list(shape_data.ear_tag_size_list),
                )

                selected_shape_color_lbl = Label(self.attr_win, text="Shape color: ")
                self.selected_shape_color = StringVar()
                self.selected_shape_color.set(current_shape_data["Color name"])
                selected_shape_color_dropdown = OptionMenu(
                    self.attr_win,
                    self.selected_shape_color,
                    *list(shape_data.named_shape_colors.keys()),
                )

                save_button = Button(
                    self.attr_win,
                    text="Save",
                    fg=shape_data.non_select_color,
                    command=lambda: self.save_attr_changes(shape_data, image_data),
                )

                attr_lbl_frame.grid(row=1, sticky=W)
                selected_shape_name_lbl.grid(row=1, column=0, sticky=W, pady=10)
                selected_shape_name_entry.grid(row=1, column=1, sticky=W, pady=10)
                selected_shape_thickness_lbl.grid(row=2, column=0, sticky=W, pady=10)
                selected_shape_thickness_dropdown.grid(
                    row=2, column=1, sticky=W, pady=10
                )
                selected_shape_eartag_size_lbl.grid(row=3, column=0, sticky=W, pady=10)
                selected_shape_eartag_size_dropdown.grid(
                    row=3, column=1, sticky=W, pady=10
                )
                selected_shape_color_lbl.grid(row=4, column=0, sticky=W, pady=10)
                selected_shape_color_dropdown.grid(row=4, column=1, sticky=W, pady=10)
                save_button.grid(row=5, column=0, sticky=W, pady=10)

            else:
                raise TypeError("No ROI selected.")

        def save_attr_changes(self, shape_data, image_data):
            new_shape_name = self.selected_shape_name_entry_txt.get()
            new_shape_thickness = self.selected_shape_thickness.get()
            new_shape_ear_tag_size = self.selected_shape_eartag_size.get()
            new_shape_color = self.selected_shape_color.get()
            for shape in [
                image_data.out_rectangles,
                image_data.out_circles,
                image_data.out_polygon,
            ]:
                for e in shape:
                    shape_type = e["Shape_type"]
                    if e["Name"] == self.shape_name:
                        e["Name"] = new_shape_name
                        e["Thickness"] = new_shape_thickness
                        e["Ear_tag_size"] = new_shape_ear_tag_size
                        e["Color name"] = new_shape_color
                        e["Color BGR"] = shape_data.named_shape_colors[new_shape_color]
                        shape_data.video_ROIs = [
                            w.replace(
                                str(shape_type) + ": " + self.shape_name,
                                str(shape_type) + ": " + new_shape_name,
                            )
                            for w in shape_data.video_ROIs
                        ]
            image_data.insert_all_ROIs_into_image()
            shape_data.update_delete_ROI_menu()
            self.attr_win.destroy()
            self.attr_win.update()

    #
    #
    def window_menus(self):
        menu = Menu(self.root)
        file_menu = Menu(menu)
        menu.add_cascade(label="File", menu=file_menu)
        file_menu.add_command(
            label="Preferences...", command=lambda: PreferenceMenu(self.image_data)
        )
        file_menu.add_separator()
        file_menu.add_command(label="Exit", command=self.Exit)
        self.root.config(menu=menu)

    def Exit(self):
        self.image_data.destroy_windows()
        self.root.destroy()


class PreferenceMenu:
    def __init__(self, image_data):
        pref_win = Toplevel()
        pref_win.minsize(400, 300)
        pref_win.wm_title("Preference Settings")
        pref_lbl_frame = LabelFrame(
            pref_win,
            text="Preferences",
            font=("Arial", 16, "bold"),
            pady=5,
            padx=5,
            fg="black",
        )
        line_type_label = Label(pref_lbl_frame, text="Shape line type: ")
        text_size_label = Label(pref_lbl_frame, text="Text size: ")
        text_thickness_label = Label(pref_lbl_frame, text="Text thickness: ")
        line_type_list = [4, 8, 16, -1]
        text_size_list = list(range(1, 20))
        text_thickness_list = list(range(1, 15))
        click_sensitivity_lbl = Label(pref_lbl_frame, text="Mouse click sensitivity: ")
        click_sensitivity_list = list(range(1, 50, 5))
        self.click_sens = IntVar()
        self.line_type = IntVar()
        self.text_size = IntVar()
        self.text_thickness = IntVar()
        self.line_type.set(line_type_list[-1])
        self.text_size.set(text_size_list[0])
        self.click_sens.set(click_sensitivity_list[0])
        line_type_dropdown = OptionMenu(pref_lbl_frame, self.line_type, *line_type_list)
        text_thickness_dropdown = OptionMenu(
            pref_lbl_frame, self.text_thickness, *text_thickness_list
        )
        text_size_dropdown = OptionMenu(pref_lbl_frame, self.text_size, *text_size_list)
        click_sens_dropdown = OptionMenu(
            pref_lbl_frame, self.click_sens, *click_sensitivity_list
        )
        duplicate_jump_size_lbl = Label(pref_lbl_frame, text="Duplicate shape jump: ")
        duplicate_jump_size_list = list(range(1, 100, 5))
        self.duplicate_jump_size = IntVar()
        self.duplicate_jump_size.set(20)
        duplicate_jump_size_dropdown = OptionMenu(
            pref_lbl_frame, self.duplicate_jump_size, *duplicate_jump_size_list
        )
        pref_save_btn = Button(
            pref_lbl_frame,
            text="Save",
            fg="black",
            command=lambda: self.save_prefs(image_data),
        )

        pref_lbl_frame.grid(row=1, sticky=W)
        line_type_label.grid(row=1, column=0, sticky=W, pady=10)
        line_type_dropdown.grid(row=1, column=1, sticky=W, pady=10)
        click_sensitivity_lbl.grid(row=2, column=0, sticky=W, pady=10)
        click_sens_dropdown.grid(row=2, column=1, sticky=W, pady=10)
        duplicate_jump_size_lbl.grid(row=3, column=0, sticky=W, pady=10)
        duplicate_jump_size_dropdown.grid(row=3, column=1, sticky=W, pady=10)
        text_size_label.grid(row=4, column=0, sticky=W, pady=10)
        text_size_dropdown.grid(row=4, column=1, sticky=W, pady=10)
        text_thickness_label.grid(row=5, column=0, sticky=W, pady=10)
        text_thickness_dropdown.grid(row=5, column=1, sticky=W, pady=10)
        pref_save_btn.grid(row=5, column=2, sticky=W, pady=10)

    def save_prefs(self, image_data):
        image_data.click_sens = self.click_sens.get()
        image_data.text_size = self.text_size.get()
        image_data.text_thickness = self.text_thickness.get()
        image_data.line_type = self.line_type.get()
        image_data.duplicate_jump_size = self.duplicate_jump_size.get()
        stdout_success(msg="Saved ROI preference settings.")


#
# test = ROI_definitions(config_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/project_config.ini',
#                       video_path='/Users/simon/Desktop/envs/simba_dev/tests/test_data/mouse_open_field/project_folder/videos/Video1.mp4')
